import asyncio
import os
from pathlib import Path
from typing import List

import click
import uvicorn
from click import Abort

from recce import event
from recce.artifact import (
    delete_dbt_artifacts,
    download_dbt_artifacts,
    upload_artifacts_to_session,
    upload_dbt_artifacts,
)
from recce.config import RECCE_CONFIG_FILE, RECCE_ERROR_LOG_FILE, RecceConfig
from recce.connect_to_cloud import (
    generate_key_pair,
    prepare_connection_url,
    run_one_time_http_server,
)
from recce.exceptions import RecceConfigException
from recce.git import current_branch, current_default_branch
from recce.run import check_github_ci_env, cli_run
from recce.server import RecceServerMode
from recce.state import (
    CloudStateLoader,
    FileStateLoader,
    RecceCloudStateManager,
    RecceShareStateManager,
)
from recce.summary import generate_markdown_summary
from recce.util.api_token import prepare_api_token, show_invalid_api_token_message
from recce.util.logger import CustomFormatter
from recce.util.onboarding_state import update_onboarding_state
from recce.util.recce_cloud import (
    RecceCloudException,
)

from .core import RecceContext, set_default_context
from .event.track import TrackCommand

event.init()


def create_state_loader(review_mode, cloud_mode, state_file, cloud_options):
    from rich.console import Console

    console = Console()

    try:
        state_loader = (
            CloudStateLoader(review_mode=review_mode, cloud_options=cloud_options)
            if cloud_mode
            else FileStateLoader(review_mode=review_mode, state_file=state_file)
        )
        state_loader.load()
        return state_loader
    except RecceCloudException as e:
        console.print("[[red]Error[/red]] Failed to load recce state file")
        console.print(f"Reason: {e.reason}")
        exit(1)
    except Exception as e:
        console.print("[[red]Error[/red]] Failed to load recce state file")
        console.print(f"Reason: {e}")
        exit(1)


def patch_derived_args(args):
    """
    Patch derived args based on other args.
    """
    if args.get("session_id") or args.get("share_url"):
        args["cloud"] = True
        args["review"] = True


def create_state_loader_by_args(state_file=None, **kwargs):
    """
    Create a state loader based on CLI arguments.

    This function handles the cloud options logic that is shared between
    server and mcp-server commands.

    Args:
        state_file: Optional path to state file
        **kwargs: CLI arguments including api_token, cloud, review, session_id, share_url, etc.

    Returns:
        state_loader: The created state loader instance
    """
    from rich.console import Console

    console = Console()

    api_token = kwargs.get("api_token")
    is_review = kwargs.get("review", False)
    is_cloud = kwargs.get("cloud", False)
    cloud_options = None

    # Handle share_url and session_id
    share_url = kwargs.get("share_url")
    session_id = kwargs.get("session_id")

    if share_url:
        share_id = share_url.split("/")[-1]
        if not share_id:
            console.print("[[red]Error[/red]] Invalid share URL format.")
            exit(1)

    if is_cloud:
        # Cloud mode
        if share_url:
            cloud_options = {
                "host": kwargs.get("state_file_host"),
                "api_token": api_token,
                "share_id": share_id,
            }
        elif session_id:
            cloud_options = {
                "host": kwargs.get("state_file_host"),
                "api_token": api_token,
                "session_id": session_id,
            }
        else:
            cloud_options = {
                "host": kwargs.get("state_file_host"),
                "github_token": kwargs.get("cloud_token"),
                "password": kwargs.get("password"),
            }

    # Create state loader
    state_loader = create_state_loader(is_review, is_cloud, state_file, cloud_options)

    return state_loader


def handle_debug_flag(**kwargs):
    if kwargs.get("debug"):
        import logging

        ch = logging.StreamHandler()
        ch.setFormatter(CustomFormatter())
        logging.basicConfig(handlers=[ch], level=logging.DEBUG)


def add_options(options):
    def _add_options(func):
        for option in reversed(options):
            func = option(func)
        return func

    return _add_options


dbt_related_options = [
    click.option("--target", "-t", help="Which target to load for the given profile.", type=click.STRING),
    click.option("--profile", help="Which existing profile to load.", type=click.STRING),
    click.option(
        "--project-dir",
        help="Which directory to look in for the dbt_project.yml file.",
        type=click.Path(),
        envvar="DBT_PROJECT_DIR",
    ),
    click.option(
        "--profiles-dir",
        help="Which directory to look in for the profiles.yml file.",
        type=click.Path(),
        envvar="DBT_PROFILES_DIR",
    ),
]

sqlmesh_related_options = [
    click.option("--sqlmesh", is_flag=True, help="Use SQLMesh ", hidden=True),
    click.option("--sqlmesh-envs", is_flag=False, help="SQLMesh envs to compare. SOURCE:TARGET", hidden=True),
    click.option("--sqlmesh-config", is_flag=False, help="SQLMesh config name to use", hidden=True),
]

recce_options = [
    click.option(
        "--config",
        help="Path to the recce config file.",
        type=click.Path(),
        default=RECCE_CONFIG_FILE,
        show_default=True,
    ),
    click.option(
        "--error-log", help="Path to the error log file.", type=click.Path(), default=RECCE_ERROR_LOG_FILE, hidden=True
    ),
    click.option("--debug", is_flag=True, help="Enable debug mode.", hidden=True),
]

recce_cloud_options = [
    click.option("--cloud", is_flag=True, help="Fetch the state file from cloud."),
    click.option(
        "--cloud-token", help="The GitHub token used by Recce Cloud.", type=click.STRING, envvar="GITHUB_TOKEN"
    ),
    click.option(
        "--state-file-host",
        help="The host to fetch the state file from.",
        type=click.STRING,
        envvar="RECCE_STATE_FILE_HOST",
        default="",
        hidden=True,
    ),
    click.option(
        "--password",
        "-p",
        help="The password to encrypt the state file in cloud.",
        type=click.STRING,
        envvar="RECCE_STATE_PASSWORD",
    ),
]

recce_cloud_auth_options = [
    click.option(
        "--api-token", help="The personal token generated by Recce Cloud.", type=click.STRING, envvar="RECCE_API_TOKEN"
    )
]

recce_dbt_artifact_dir_options = [
    click.option(
        "--target-path",
        help="dbt artifacts directory for your development branch.",
        type=click.STRING,
        default="target",
    ),
    click.option(
        "--target-base-path",
        help="dbt artifacts directory to be used as the base for the comparison.",
        type=click.STRING,
        default="target-base",
    ),
]

recce_hidden_options = [
    click.option(
        "--mode",
        envvar="RECCE_SERVER_MODE",
        type=click.Choice(RecceServerMode.available_members(), case_sensitive=False),
        hidden=True,
    ),
    click.option(
        "--share-url",
        help="The share URL triggers this instance.",
        type=click.STRING,
        envvar="RECCE_SHARE_URL",
        hidden=True,
    ),
    click.option(
        "--session-id",
        help="The session ID triggers this instance.",
        type=click.STRING,
        envvar=["RECCE_SESSION_ID", "RECCE_SNAPSHOT_ID"],  # Backward compatibility with RECCE_SNAPSHOT_ID
        hidden=True,
    ),
]


def _execute_sql(context, sql_template, base=False):
    try:
        import pandas as pd
    except ImportError:
        print("'pandas' package not found. You can install it using the command: 'pip install pandas'.")
        exit(1)

    from recce.adapter.dbt_adapter import DbtAdapter

    dbt_adapter: DbtAdapter = context.adapter
    with dbt_adapter.connection_named("recce"):
        sql = dbt_adapter.generate_sql(sql_template, base)
        response, result = dbt_adapter.execute(sql, fetch=True, auto_begin=True)
        table = result
        df = pd.DataFrame([row.values() for row in table.rows], columns=table.column_names)
        return df


@click.group()
@click.pass_context
def cli(ctx, **kwargs):
    """Recce: Data validation toolkit for comprehensive PR review"""
    from rich.console import Console

    from recce import __is_recce_outdated__, __latest_version__

    if __is_recce_outdated__ is True:
        error_console = Console(stderr=True, style="bold")
        error_console.print(
            f"[[yellow]Update Available[/yellow]] A new version of Recce {__latest_version__} is available.",
        )
        error_console.print("Please update using the command: 'pip install --upgrade recce'.", end="\n\n")


@cli.command(cls=TrackCommand)
def version():
    """
    Show version information
    """
    from recce import __version__

    print(__version__)


@cli.command(cls=TrackCommand)
@add_options(dbt_related_options)
@add_options(recce_dbt_artifact_dir_options)
def debug(**kwargs):
    """
    Diagnose and verify Recce setup for the development and the base environments
    """

    from rich.console import Console

    from recce.adapter.dbt_adapter import DbtAdapter
    from recce.core import load_context

    console = Console()

    def check_artifacts(env_name, target_path):
        console.rule(f"{env_name} Environment", style="orange3")
        if not target_path.is_dir():
            console.print(f"[[red]MISS[/red]] Directory not found: {target_path}")
            return [False, False, False]

        console.print(f"[[green]OK[/green]] Directory exists: {target_path}")

        manifest_path = target_path / "manifest.json"
        manifest_is_ready = manifest_path.is_file()
        if manifest_is_ready:
            console.print(f"[[green]OK[/green]] Manifest JSON file exists : {manifest_path}")
        else:
            console.print(f"[[red]MISS[/red]] Manifest JSON file not found: {manifest_path}")

        catalog_path = target_path / "catalog.json"
        catalog_is_ready = catalog_path.is_file()
        if catalog_is_ready:
            console.print(f"[[green]OK[/green]] Catalog JSON file exists: {catalog_path}")
        else:
            console.print(f"[[red]MISS[/red]] Catalog JSON file not found: {catalog_path}")

        return [True, manifest_is_ready, catalog_is_ready]

    project_dir_path = Path(kwargs.get("project_dir") or "./")
    target_path = project_dir_path.joinpath(Path(kwargs.get("target_path", "target")))
    target_base_path = project_dir_path.joinpath(Path(kwargs.get("target_base_path", "target-base")))

    curr_is_ready = check_artifacts("Development", target_path)
    base_is_ready = check_artifacts("Base", target_base_path)

    console.rule("Warehouse Connection", style="orange3")
    conn_is_ready = True
    try:
        context_kwargs = {**kwargs, "target_base_path": kwargs.get("target_path")}
        ctx = load_context(**context_kwargs)
        dbt_adapter: DbtAdapter = ctx.adapter
        sql = dbt_adapter.generate_sql("select 1", False)
        dbt_adapter.execute(sql, fetch=True, auto_begin=True)
        console.print("[[green]OK[/green]] Connection test")
    except Exception:
        conn_is_ready = False
        console.print("[[red]FAIL[/red]] Connection test")

    console.rule("Result", style="orange3")
    if all(curr_is_ready) and all(base_is_ready) and conn_is_ready:
        console.print("[[green]OK[/green]] Ready to launch! Type 'recce server'.")
    elif all(curr_is_ready) and conn_is_ready:
        console.print("[[orange3]OK[/orange3]] Ready to launch with [i]limited features[/i]. Type 'recce server'.")

    if not curr_is_ready[0]:
        console.print(
            "[[orange3]TIP[/orange3]] Run dbt or overwrite the default directory of the development environment with '--target-path'."
        )
    else:
        if not curr_is_ready[1]:
            console.print(
                "[[orange3]TIP[/orange3]] 'dbt run' to generate the manifest JSON file for the development environment."
            )
        if not curr_is_ready[2]:
            console.print(
                "[[orange3]TIP[/orange3]] 'dbt docs generate' to generate the catalog JSON file for the development environment."
            )

    if not base_is_ready[0]:
        console.print(
            "[[orange3]TIP[/orange3]] Run dbt with '--target-path target-base' or overwrite the default directory of the base environment with '--target-base-path'."
        )
    else:
        if not base_is_ready[1]:
            console.print(
                "[[orange3]TIP[/orange3]] 'dbt run --target-path target-base' to generate the manifest JSON file for the base environment."
            )
        if not base_is_ready[2]:
            console.print(
                "[[orange3]TIP[/orange3]] 'dbt docs generate --target-path target-base' to generate the catalog JSON file for the base environment."
            )

    if not conn_is_ready:
        console.print("[[orange3]TIP[/orange3]] Run 'dbt debug' to check the connection.")


@cli.command(hidden=True, cls=TrackCommand)
@click.option("--sql", help="Sql template to query", required=True)
@click.option("--base", is_flag=True, help="Run the query on the base environment")
@add_options(dbt_related_options)
def query(sql, base: bool = False, **kwargs):
    """
    Run a query on the current or base environment

    Examples:\n

    - run an adhoc query\n
        recce query --sql 'select * from {{ ref("mymodel") }} order by 1'

    - run an adhoc query on base environment\n
        recce query --base --sql 'select * from {{ ref("mymodel") }} order by 1'
    """
    context = RecceContext.load(**kwargs)
    result = _execute_sql(context, sql, base=base)
    print(result.to_string(na_rep="-", index=False))


def _split_comma_separated(ctx, param, value):
    return value.split(",") if value else None


@cli.command(hidden=True, cls=TrackCommand)
@click.option("--sql", help="Sql template to query.", required=True)
@click.option(
    "--primary-keys",
    type=click.STRING,
    help="Comma-separated list of primary key columns.",
    callback=_split_comma_separated,
)
@click.option("--keep-shape", is_flag=True, help="Keep unchanged columns. Otherwise, unchanged columns are hidden.")
@click.option(
    "--keep-equal", is_flag=True, help='Keep values that are equal. Otherwise, equal values are shown as "-".'
)
@add_options(dbt_related_options)
def diff(sql, primary_keys: List[str] = None, keep_shape: bool = False, keep_equal: bool = False, **kwargs):
    """
    Run queries on base and current environments and diff the results

    Examples:\n

    - run adhoc queries and diff the results\n
        recce diff --sql 'select * from {{ ref("mymodel") }} order by 1'
    """

    context = RecceContext.load(**kwargs)
    before = _execute_sql(context, sql, base=True)
    if primary_keys is not None:
        before.set_index(primary_keys, inplace=True)
    after = _execute_sql(context, sql, base=False)
    if primary_keys is not None:
        after.set_index(primary_keys, inplace=True)

    before_aligned, after_aligned = before.align(after)
    diff = before_aligned.compare(
        after_aligned, result_names=("base", "current"), keep_equal=keep_equal, keep_shape=keep_shape
    )
    print(diff.to_string(na_rep="-") if not diff.empty else "no changes")


@cli.command(cls=TrackCommand)
@click.argument("state_file", required=False)
@click.option("--host", default="localhost", show_default=True, help="The host to bind to.")
@click.option("--port", default=8000, show_default=True, help="The port to bind to.", type=int)
@click.option("--lifetime", default=0, show_default=True, help="The lifetime of the server in seconds.", type=int)
@click.option("--review", is_flag=True, help="Open the state file in the review mode.")
@click.option("--single-env", is_flag=True, help="Launch in single environment mode directly.")
@add_options(dbt_related_options)
@add_options(sqlmesh_related_options)
@add_options(recce_options)
@add_options(recce_dbt_artifact_dir_options)
@add_options(recce_cloud_options)
@add_options(recce_cloud_auth_options)
@add_options(recce_hidden_options)
def server(host, port, lifetime, state_file=None, **kwargs):
    """
    Launch the recce server

    STATE_FILE is the path to the recce state file. Defaults=None, which will be no persistent state.

    Examples:\n

    \b
    # Launch the recce server
    recce server

    \b
    # Launch the recce server with a state file
    recce server recce_state.json

    \b
    # Launch the server in the review mode
    recce server --review recce_state.json

    \b
    # Launch the server using the state from the PR of your current branch. (Requires GitHub token)
    export GITHUB_TOKEN=<your-github-token>
    recce server --cloud
    recce server --review --cloud

    """

    from rich.console import Console
    from rich.prompt import Confirm

    from .server import AppState, app

    RecceConfig(config_file=kwargs.get("config"))

    handle_debug_flag(**kwargs)
    patch_derived_args(kwargs)

    server_mode = kwargs.get("mode") if kwargs.get("mode") else RecceServerMode.server
    is_review = kwargs.get("review", False)
    is_cloud = kwargs.get("cloud", False)
    flag = {
        "single_env_onboarding": False,
        "show_relaunch_hint": False,
        "preview": False,
        "read_only": False,
    }
    console = Console()

    # Prepare API token
    try:
        api_token = prepare_api_token(**kwargs)
        kwargs["api_token"] = api_token
    except RecceConfigException:
        show_invalid_api_token_message()
        exit(1)
    auth_options = {
        "api_token": api_token,
    }

    # Check Single Environment Onboarding Mode if not in cloud mode and not in review mode
    if not is_cloud and not is_review:
        project_dir_path = Path(kwargs.get("project_dir") or "./")
        target_base_path = project_dir_path.joinpath(Path(kwargs.get("target_base_path", "target-base")))
        if not target_base_path.is_dir():
            # Mark as single env onboarding mode if user provides the target-path only
            flag["single_env_onboarding"] = True
            flag["show_relaunch_hint"] = True
            # Use the target path as the base path
            kwargs["target_base_path"] = kwargs.get("target_path")

    # Server mode:
    #
    # It's used to determine the features disabled in the Web UI. Only used in the cloud-managed recce instances.
    #
    # Read-Only: No run query, no checklist
    # Preview (Metadata-Only): No run query
    if server_mode == RecceServerMode.preview:
        flag["preview"] = True
    elif server_mode == RecceServerMode.read_only:
        flag["read_only"] = True

    # Onboarding State logic update here
    update_onboarding_state(api_token, flag.get("single_env_onboarding"))

    # Create state loader using shared function
    state_loader = create_state_loader_by_args(state_file, **kwargs)

    if not state_loader.verify():
        error, hint = state_loader.error_and_hint
        console.print(f"[[red]Error[/red]] {error}")
        console.print(f"{hint}")
        exit(1)

    try:
        result, message = RecceContext.verify_required_artifacts(**kwargs)
    except Exception as e:
        result = False
        error_type = type(e).__name__
        error_message = str(e)
        message = f"{error_type}: {error_message}"
    if not result:
        console.rule("Notice", style="orange3")
        console.print(f"[[red]Error[/red]] {message}")
        exit(1)

    if state_loader.review_mode:
        console.rule("Recce Server : Review Mode")
    elif flag.get("single_env_onboarding"):
        # Show warning message
        console.rule("Notice", style="orange3")
        console.print(
            "Recce will launch with limited features (no environment comparison).\n"
            "\n"
            "For full functionality, set up a base environment first.\n"
            "Setup help: 'recce debug' or https://docs.datarecce.io/configure-diff/\n"
        )

        single_env_flag = kwargs.get("single_env", False)
        if not single_env_flag:
            lanch_in_single_env = Confirm.ask("Continue to launch Recce?")
            if not lanch_in_single_env:
                exit(0)

        console.rule("Recce Server : Limited Features")
    else:
        console.rule("Recce Server")

    state = AppState(
        command=server_mode,
        state_loader=state_loader,
        kwargs=kwargs,
        flag=flag,
        auth_options=auth_options,
        lifetime=lifetime,
        share_url=kwargs.get("share_url"),
    )
    app.state = state

    if server_mode == RecceServerMode.read_only:
        set_default_context(RecceContext.load(**kwargs, state_loader=state_loader))

    uvicorn.run(app, host=host, port=port, lifespan="on")


DEFAULT_RECCE_STATE_FILE = "recce_state.json"


@cli.command(cls=TrackCommand)
@click.option(
    "-o",
    "--output",
    help="Path of the output state file.",
    type=click.Path(),
    default=DEFAULT_RECCE_STATE_FILE,
    show_default=True,
)
@click.option("--state-file", help="Path of the import state file.", type=click.Path())
@click.option("--summary", help="Path of the summary markdown file.", type=click.Path())
@click.option("--skip-query", is_flag=True, help="Skip running the queries for the checks.")
@click.option("--skip-check", is_flag=True, help="Skip running the checks.")
@click.option(
    "--git-current-branch",
    help="The git branch of the current environment.",
    type=click.STRING,
    envvar="GITHUB_HEAD_REF",
)
@click.option(
    "--git-base-branch", help="The git branch of the base environment.", type=click.STRING, envvar="GITHUB_BASE_REF"
)
@click.option(
    "--github-pull-request-url", help="The github pull request url to use for the lineage.", type=click.STRING
)
@add_options(dbt_related_options)
@add_options(sqlmesh_related_options)
@add_options(recce_options)
@add_options(recce_dbt_artifact_dir_options)
@add_options(recce_cloud_options)
def run(output, **kwargs):
    """
    Run recce and output the state file

    Examples:\n

    \b
    # Run recce and output to the default path [recce_state.json]
    recce run

    \b
    # Run recce and output to the specified path
    recce run -o my_recce_state.json

    \b
    # Run recce and output to the specified path
    recce run --cloud --cloud-token <token> --password <password>

    """
    from rich.console import Console

    handle_debug_flag(**kwargs)
    console = Console()
    is_github_action, pr_url = check_github_ci_env(**kwargs)
    if is_github_action is True and pr_url is not None:
        kwargs["github_pull_request_url"] = pr_url

    # Initialize Recce Config
    RecceConfig(config_file=kwargs.get("config"))

    cloud_mode = kwargs.get("cloud", False)
    state_file = kwargs.get("state_file")
    cloud_options = (
        {
            "host": kwargs.get("state_file_host"),
            "github_token": kwargs.get("cloud_token"),
            "password": kwargs.get("password"),
        }
        if cloud_mode
        else None
    )

    state_loader = create_state_loader(
        review_mode=False, cloud_mode=cloud_mode, state_file=state_file, cloud_options=cloud_options
    )

    if not state_loader.verify():
        error, hint = state_loader.error_and_hint
        console.print(f"[[red]Error[/red]] {error}")
        console.print(f"{hint}")
        exit(1)

    result, message = RecceContext.verify_required_artifacts(**kwargs)
    if not result:
        console.print(f"[[red]Error[/red]] {message}")
        exit(1)

    # Verify the output state file path
    try:
        if os.path.isdir(output) or output.endswith("/"):

            output_dir = Path(output)
            # Create the directory if not exists
            output_dir.mkdir(parents=True, exist_ok=True)
            output = os.path.join(output, DEFAULT_RECCE_STATE_FILE)
            console.print(
                f"[[yellow]Warning[/yellow]] The path '{output_dir}' is a directory. "
                f"The state file will be saved as '{output}'."
            )
        else:
            # Create the parent directory if not exists
            output_dir = Path(output).parent
            output_dir.mkdir(parents=True, exist_ok=True)
    except FileExistsError as e:
        console.print(f"[[red]Error[/red]] Failed to access file path '{output}'.")
        console.print(f"Reason: {e}")
        exit(1)

    return asyncio.run(cli_run(output, state_loader=state_loader, **kwargs))


@cli.command(cls=TrackCommand)
@click.argument("state_file", required=False)
@click.option(
    "--format",
    "-f",
    help="Output format. Currently only markdown is supported.",
    type=click.Choice(["markdown", "mermaid", "check"], case_sensitive=False),
    default="markdown",
    show_default=True,
    hidden=True,
)
@add_options(dbt_related_options)
@add_options(recce_options)
@add_options(recce_cloud_options)
def summary(state_file, **kwargs):
    """
    Generate a summary of the recce state file
    """
    from rich.console import Console

    from .core import load_context

    handle_debug_flag(**kwargs)
    console = Console()
    cloud_mode = kwargs.get("cloud", False)
    cloud_options = (
        {
            "host": kwargs.get("state_file_host"),
            "github_token": kwargs.get("cloud_token"),
            "password": kwargs.get("password"),
        }
        if cloud_mode
        else None
    )

    state_loader = create_state_loader(
        review_mode=True, cloud_mode=cloud_mode, state_file=state_file, cloud_options=cloud_options
    )

    if not state_loader.verify():
        error, hint = state_loader.error_and_hint
        console.print(f"[[red]Error[/red]] {error}")
        console.print(f"{hint}")
        exit(1)
    try:
        # Load context in review mode, won't need to check dbt_project.yml file.
        ctx = load_context(**kwargs, state_loader=state_loader, review=True)
    except Exception as e:
        console.print("[[red]Error[/red]] Failed to generate summary")
        console.print(f"{e}")
        exit(1)

    output = generate_markdown_summary(ctx, summary_format=kwargs.get("format"))
    print(output)


@cli.command(cls=TrackCommand)
def connect_to_cloud():
    """
    Connect OSS to Cloud
    """
    import webbrowser

    from rich.console import Console

    console = Console()

    # Prepare RSA keys for connecting to cloud
    private_key, public_key = generate_key_pair()

    connect_url, callback_port = prepare_connection_url(public_key)
    console.rule("Connecting to Recce Cloud")
    console.print("Attempting to automatically open the Recce Cloud authorization page in your default browser.")
    console.print("If the browser does not open, please open the following URL:")
    console.print(connect_url)
    webbrowser.open(connect_url)

    # Launch a callback HTTP server for fetching the api-token
    run_one_time_http_server(private_key, port=callback_port)


@cli.group("cloud", short_help="Manage Recce Cloud state file.")
def cloud(**kwargs):
    # Manage Recce Cloud.
    pass


@cloud.command(cls=TrackCommand)
@click.option("--cloud-token", help="The GitHub token used by Recce Cloud.", type=click.STRING, envvar="GITHUB_TOKEN")
@click.option(
    "--state-file-host",
    help="The host to fetch the state file from.",
    type=click.STRING,
    envvar="RECCE_STATE_FILE_HOST",
    default="",
    hidden=True,
)
@click.option(
    "--password",
    "-p",
    help="The password to encrypt the state file in cloud.",
    type=click.STRING,
    envvar="RECCE_STATE_PASSWORD",
)
@click.option("--force", "-f", help="Bypasses the confirmation prompt. Purge the state file directly.", is_flag=True)
@add_options(recce_options)
def purge(**kwargs):
    """
    Purge the state file from cloud
    """
    from rich.console import Console

    handle_debug_flag(**kwargs)
    console = Console()
    state_loader = None
    cloud_options = {
        "host": kwargs.get("state_file_host"),
        "github_token": kwargs.get("cloud_token"),
        "password": kwargs.get("password"),
    }
    force_to_purge = kwargs.get("force", False)

    try:
        console.rule("Check Recce State from Cloud")
        state_loader = create_state_loader(
            review_mode=False, cloud_mode=True, state_file=None, cloud_options=cloud_options
        )
    except Exception:
        console.print("[[yellow]Skip[/yellow]] Cannot access existing state file from cloud. Purge it directly.")

    if state_loader is None:
        try:
            if force_to_purge is True or click.confirm("\nDo you want to purge the state file?"):
                rc, err_msg = RecceCloudStateManager(cloud_options).purge_cloud_state()
                if rc is True:
                    console.rule("Purged Successfully")
                else:
                    console.rule("Failed to Purge", style="red")
                    console.print(f"Reason: {err_msg}")

        except click.exceptions.Abort:
            pass
        return 0

    info = state_loader.info()
    if info is None:
        console.print("[[yellow]Skip[/yellow]] No state file found in cloud.")
        return 0

    pr_info = info.get("pull_request")
    console.print("[green]State File hosted by[/green]", info.get("source"))
    console.print("[green]GitHub Repository[/green]", info.get("pull_request").repository)
    console.print(f"[green]GitHub Pull Request[/green]\n{pr_info.title} #{pr_info.id}")
    console.print(f"Branch merged into [blue]{pr_info.base_branch}[/blue] from [blue]{pr_info.branch}[/blue]")
    console.print(pr_info.url)

    try:
        if force_to_purge is True or click.confirm("\nDo you want to purge the state file?"):
            response = state_loader.purge()
            if response is True:
                console.rule("Purged Successfully")
            else:
                console.rule("Failed to Purge", style="red")
                console.print(f"Reason: {state_loader.error_message}")
    except click.exceptions.Abort:
        pass

    return 0


@cloud.command(cls=TrackCommand)
@click.argument("state_file", type=click.Path(exists=True))
@click.option("--cloud-token", help="The GitHub token used by Recce Cloud.", type=click.STRING, envvar="GITHUB_TOKEN")
@click.option(
    "--state-file-host",
    help="The host to fetch the state file from.",
    type=click.STRING,
    envvar="RECCE_STATE_FILE_HOST",
    default="",
    hidden=True,
)
@click.option(
    "--password",
    "-p",
    help="The password to encrypt the state file in cloud.",
    type=click.STRING,
    envvar="RECCE_STATE_PASSWORD",
)
@add_options(recce_options)
def upload(state_file, **kwargs):
    """
    Upload the state file to cloud
    """
    from rich.console import Console

    handle_debug_flag(**kwargs)
    cloud_options = {
        "host": kwargs.get("state_file_host"),
        "github_token": kwargs.get("cloud_token"),
        "password": kwargs.get("password"),
    }

    console = Console()

    # load local state
    state_loader = create_state_loader(
        review_mode=False, cloud_mode=False, state_file=state_file, cloud_options=cloud_options
    )

    if not state_loader.verify():
        error, hint = state_loader.error_and_hint
        console.print(f"[[red]Error[/red]] {error}")
        console.print(f"{hint}")
        exit(1)

    # check if state exists in cloud
    state_manager = RecceCloudStateManager(cloud_options)
    if not state_manager.verify():
        error, hint = state_manager.error_and_hint
        console.print(f"[[red]Error[/red]] {error}")
        console.print(f"{hint}")
        exit(1)

    cloud_state_file_exists = state_manager.check_cloud_state_exists()

    if cloud_state_file_exists and not click.confirm("\nDo you want to overwrite the existing state file?"):
        return 0

    console.print(state_manager.upload_state_to_cloud(state_loader.state))


@cloud.command(cls=TrackCommand)
@click.option(
    "-o",
    "--output",
    help="Path of the downloaded state file.",
    type=click.STRING,
    default=DEFAULT_RECCE_STATE_FILE,
    show_default=True,
)
@click.option("--cloud-token", help="The GitHub token used by Recce Cloud.", type=click.STRING, envvar="GITHUB_TOKEN")
@click.option(
    "--state-file-host",
    help="The host to fetch the state file from.",
    type=click.STRING,
    envvar="RECCE_STATE_FILE_HOST",
    default="",
    hidden=True,
)
@click.option(
    "--password",
    "-p",
    help="The password to encrypt the state file in cloud.",
    type=click.STRING,
    envvar="RECCE_STATE_PASSWORD",
)
@add_options(recce_options)
def download(**kwargs):
    """
    Download the state file to cloud
    """
    from rich.console import Console

    handle_debug_flag(**kwargs)
    filepath = kwargs.get("output")
    cloud_options = {
        "host": kwargs.get("state_file_host"),
        "github_token": kwargs.get("cloud_token"),
        "password": kwargs.get("password"),
    }

    console = Console()

    # check if state exists in cloud
    state_manager = RecceCloudStateManager(cloud_options)
    if not state_manager.verify():
        error, hint = state_manager.error_and_hint
        console.print(f"[[red]Error[/red]] {error}")
        console.print(f"{hint}")
        exit(1)

    cloud_state_file_exists = state_manager.check_cloud_state_exists()

    if not cloud_state_file_exists:
        console.print("[yellow]Skip[/yellow] No state file found in cloud.")
        return 0

    state_manager.download_state_from_cloud(filepath)
    console.print(f'Downloaded state file to "{filepath}"')


@cloud.command(cls=TrackCommand)
@click.option("--cloud-token", help="The GitHub token used by Recce Cloud.", type=click.STRING, envvar="GITHUB_TOKEN")
@click.option(
    "--branch",
    "-b",
    help="The branch of the provided artifacts.",
    type=click.STRING,
    envvar="GITHUB_HEAD_REF",
    default=current_branch(),
    show_default=True,
)
@click.option(
    "--target-path",
    help="dbt artifacts directory for your artifacts.",
    type=click.STRING,
    default="target",
    show_default=True,
)
@click.option(
    "--password",
    "-p",
    help="The password to encrypt the dbt artifacts in cloud.",
    type=click.STRING,
    envvar="RECCE_STATE_PASSWORD",
    required=True,
)
@add_options(recce_options)
def upload_artifacts(**kwargs):
    """
    Upload the dbt artifacts to cloud

    Upload the dbt artifacts (metadata.json, catalog.json) to Recce Cloud for the given branch.
    The password is used to encrypt the dbt artifacts in the cloud. You will need the password to download the dbt artifacts.

    By default, the artifacts are uploaded to the current branch. You can specify the branch using the --branch option.
    The target path is set to 'target' by default. You can specify the target path using the --target-path option.
    """
    from rich.console import Console

    console = Console()
    cloud_token = kwargs.get("cloud_token")
    password = kwargs.get("password")
    target_path = kwargs.get("target_path")
    branch = kwargs.get("branch")

    try:
        rc = upload_dbt_artifacts(
            target_path, branch=branch, token=cloud_token, password=password, debug=kwargs.get("debug", False)
        )
        console.rule("Uploaded Successfully")
        console.print(
            f'Uploaded dbt artifacts to Recce Cloud for branch "{branch}" from "{os.path.abspath(target_path)}"'
        )
    except Exception as e:
        console.rule("Failed to Upload", style="red")
        console.print("[[red]Error[/red]] Failed to upload the dbt artifacts to cloud.")
        console.print(f"Reason: {e}")
        rc = 1
    return rc


def _download_artifacts(branch, cloud_token, console, kwargs, password, target_path):
    try:
        rc = download_dbt_artifacts(
            target_path,
            branch=branch,
            token=cloud_token,
            password=password,
            force=kwargs.get("force", False),
            debug=kwargs.get("debug", False),
        )
        console.rule("Downloaded Successfully")
        console.print(
            f'Downloaded dbt artifacts from Recce Cloud for branch "{branch}" to "{os.path.abspath(target_path)}"'
        )
    except Exception as e:
        console.rule("Failed to Download", style="red")
        console.print("[[red]Error[/red]] Failed to download the dbt artifacts from cloud.")
        reason = str(e)

        if (
            "Requests specifying Server Side Encryption with Customer provided keys must provide the correct secret key"
            in reason
        ):
            console.print("Reason: Decryption failed due to incorrect password.")
            console.print(
                "Please provide the correct password to decrypt the dbt artifacts. Or re-upload the dbt artifacts with a new password."
            )
        elif "The specified key does not exist" in reason:
            console.print("Reason: The dbt artifacts is not found in the cloud.")
            console.print("Please upload the dbt artifacts to the cloud before downloading it.")
        else:
            console.print(f"Reason: {reason}")
        rc = 1
    return rc


@cloud.command(cls=TrackCommand)
@click.option("--cloud-token", help="The GitHub token used by Recce Cloud.", type=click.STRING, envvar="GITHUB_TOKEN")
@click.option(
    "--branch",
    "-b",
    help="The branch of the selected artifacts.",
    type=click.STRING,
    envvar="GITHUB_BASE_REF",
    default=current_branch(),
    show_default=True,
)
@click.option(
    "--target-path",
    help="The dbt artifacts directory for your artifacts.",
    type=click.STRING,
    default="target",
    show_default=True,
)
@click.option(
    "--password",
    "-p",
    help="The password to decrypt the dbt artifacts in cloud.",
    type=click.STRING,
    envvar="RECCE_STATE_PASSWORD",
    required=True,
)
@click.option("--force", "-f", help="Bypasses the confirmation prompt. Download the artifacts directly.", is_flag=True)
@add_options(recce_options)
def download_artifacts(**kwargs):
    """
    Download the dbt artifacts from cloud

    Download the dbt artifacts (metadata.json, catalog.json) from Recce Cloud for the given branch.
    The password is used to decrypt the dbt artifacts in the cloud.

    By default, the artifacts are downloaded from the current branch. You can specify the branch using the --branch option.
    The target path is set to 'target' by default. You can specify the target path using the --target-path option.
    """
    from rich.console import Console

    console = Console()
    cloud_token = kwargs.get("cloud_token")
    password = kwargs.get("password")
    target_path = kwargs.get("target_path")
    branch = kwargs.get("branch")
    return _download_artifacts(branch, cloud_token, console, kwargs, password, target_path)


@cloud.command(cls=TrackCommand)
@click.option("--cloud-token", help="The GitHub token used by Recce Cloud.", type=click.STRING, envvar="GITHUB_TOKEN")
@click.option(
    "--branch",
    "-b",
    help="The branch of the selected artifacts.",
    type=click.STRING,
    envvar="GITHUB_BASE_REF",
    default=current_default_branch(),
    show_default=True,
)
@click.option(
    "--target-path",
    help="The dbt artifacts directory for your artifacts.",
    type=click.STRING,
    default="target-base",
    show_default=True,
)
@click.option(
    "--password",
    "-p",
    help="The password to decrypt the dbt artifacts in cloud.",
    type=click.STRING,
    envvar="RECCE_STATE_PASSWORD",
    required=True,
)
@click.option("--force", "-f", help="Bypasses the confirmation prompt. Download the artifacts directly.", is_flag=True)
@add_options(recce_options)
def download_base_artifacts(**kwargs):
    """
    Download the base dbt artifacts from cloud

    Download the base dbt artifacts (metadata.json, catalog.json) from Recce Cloud.
    This is useful when you start to set up the base dbt artifacts for the first time.

    Please make sure you have uploaded the dbt artifacts before downloading them.
    """
    from rich.console import Console

    console = Console()
    cloud_token = kwargs.get("cloud_token")
    password = kwargs.get("password")
    target_path = kwargs.get("target_path")
    branch = kwargs.get("branch")
    # If recce can't infer default branch from "GITHUB_BASE_REF" and current_default_branch()
    if branch is None:
        console.print(
            "[[red]Error[/red]] Please provide your base branch name with '--branch' to download the base " "artifacts."
        )
        exit(1)

    return _download_artifacts(branch, cloud_token, console, kwargs, password, target_path)


@cloud.command(cls=TrackCommand)
@click.option("--cloud-token", help="The GitHub token used by Recce Cloud.", type=click.STRING, envvar="GITHUB_TOKEN")
@click.option(
    "--branch",
    "-b",
    help="The branch to delete artifacts from.",
    type=click.STRING,
    envvar="GITHUB_HEAD_REF",
    default=current_branch(),
    show_default=True,
)
@click.option("--force", "-f", help="Bypasses the confirmation prompt. Delete the artifacts directly.", is_flag=True)
@add_options(recce_options)
def delete_artifacts(**kwargs):
    """
    Delete the dbt artifacts from cloud

    Delete the dbt artifacts (metadata.json, catalog.json) from Recce Cloud for the given branch.
    This will permanently remove the artifacts from the cloud storage.

    By default, the artifacts are deleted from the current branch. You can specify the branch using the --branch option.
    """
    from rich.console import Console

    console = Console()
    cloud_token = kwargs.get("cloud_token")
    branch = kwargs.get("branch")
    force = kwargs.get("force", False)

    if not force:
        if not click.confirm(f'Do you want to delete artifacts from branch "{branch}"?'):
            console.print("Deletion cancelled.")
            return 0

    try:
        delete_dbt_artifacts(branch=branch, token=cloud_token, debug=kwargs.get("debug", False))
        console.print(f"[[green]Success[/green]] Artifacts deleted from branch: {branch}")
        return 0
    except click.exceptions.Abort:
        pass
    except RecceCloudException as e:
        console.print("[[red]Error[/red]] Failed to delete the dbt artifacts from cloud.")
        console.print(f"Reason: {e.reason}")
        exit(1)
    except Exception as e:
        console.print("[[red]Error[/red]] Failed to delete the dbt artifacts from cloud.")
        console.print(f"Reason: {e}")
        exit(1)


@cloud.command(cls=TrackCommand, name="list-organizations")
@click.option("--api-token", help="The Recce Cloud API token.", type=click.STRING, envvar="RECCE_CLOUD_API_TOKEN")
@add_options(recce_options)
def list_organizations(**kwargs):
    """
    List organizations from Recce Cloud

    Lists all organizations that the authenticated user has access to.
    """
    from rich.console import Console
    from rich.table import Table

    console = Console()
    handle_debug_flag(**kwargs)

    try:
        api_token = prepare_api_token(**kwargs)
    except RecceConfigException:
        show_invalid_api_token_message()
        exit(1)

    try:
        from recce.util.recce_cloud import RecceCloud

        cloud = RecceCloud(api_token)
        organizations = cloud.list_organizations()

        if not organizations:
            console.print("No organizations found.")
            return

        table = Table(title="Organizations")
        table.add_column("ID", style="cyan")
        table.add_column("Name", style="green")
        table.add_column("Display Name", style="yellow")

        for org in organizations:
            table.add_row(str(org.get("id", "")), org.get("name", ""), org.get("display_name", ""))

        console.print(table)

    except RecceCloudException as e:
        console.print(f"[[red]Error[/red]] {e}")
        exit(1)
    except Exception as e:
        console.print(f"[[red]Error[/red]] {e}")
        exit(1)


@cloud.command(cls=TrackCommand, name="list-projects")
@click.option(
    "--organization",
    "-o",
    help="Organization ID (can also be set via RECCE_ORGANIZATION_ID environment variable)",
    type=click.STRING,
    envvar="RECCE_ORGANIZATION_ID",
)
@click.option("--api-token", help="The Recce Cloud API token.", type=click.STRING, envvar="RECCE_CLOUD_API_TOKEN")
@add_options(recce_options)
def list_projects(**kwargs):
    """
    List projects from Recce Cloud

    Lists all projects in the specified organization that the authenticated user has access to.

    Examples:

        # Using environment variable
        export RECCE_ORGANIZATION_ID=8
        recce cloud list-projects

        # Using command line argument
        recce cloud list-projects --organization 8

        # Override environment variable
        export RECCE_ORGANIZATION_ID=8
        recce cloud list-projects --organization 10
    """
    from rich.console import Console
    from rich.table import Table

    console = Console()
    handle_debug_flag(**kwargs)

    try:
        api_token = prepare_api_token(**kwargs)
    except RecceConfigException:
        show_invalid_api_token_message()
        exit(1)

    organization = kwargs.get("organization")
    if not organization:
        console.print("[[red]Error[/red]] Organization ID is required. Please provide it via:")
        console.print("  --organization <id> or set RECCE_ORGANIZATION_ID environment variable")
        exit(1)

    try:
        from recce.util.recce_cloud import RecceCloud

        cloud = RecceCloud(api_token)
        projects = cloud.list_projects(organization)

        if not projects:
            console.print(f"No projects found in organization {organization}.")
            return

        table = Table(title=f"Projects in Organization {organization}")
        table.add_column("ID", style="cyan")
        table.add_column("Name", style="green")
        table.add_column("Display Name", style="yellow")

        for project in projects:
            table.add_row(str(project.get("id", "")), project.get("name", ""), project.get("display_name", ""))

        console.print(table)

    except RecceCloudException as e:
        console.print(f"[[red]Error[/red]] {e}")
        exit(1)
    except Exception as e:
        console.print(f"[[red]Error[/red]] {e}")
        exit(1)


@cloud.command(cls=TrackCommand, name="list-sessions")
@click.option(
    "--organization",
    "-o",
    help="Organization ID (can also be set via RECCE_ORGANIZATION_ID environment variable)",
    type=click.STRING,
    envvar="RECCE_ORGANIZATION_ID",
)
@click.option(
    "--project",
    "-p",
    help="Project ID (can also be set via RECCE_PROJECT_ID environment variable)",
    type=click.STRING,
    envvar="RECCE_PROJECT_ID",
)
@click.option("--api-token", help="The Recce Cloud API token.", type=click.STRING, envvar="RECCE_CLOUD_API_TOKEN")
@add_options(recce_options)
def list_sessions(**kwargs):
    """
    List sessions from Recce Cloud

    Lists all sessions in the specified project that the authenticated user has access to.

    Examples:

        # Using environment variables
        export RECCE_ORGANIZATION_ID=8
        export RECCE_PROJECT_ID=7
        recce cloud list-sessions

        # Using command line arguments
        recce cloud list-sessions --organization 8 --project 7

        # Mixed usage (env + CLI override)
        export RECCE_ORGANIZATION_ID=8
        recce cloud list-sessions --project 7

        # Override environment variables
        export RECCE_ORGANIZATION_ID=8
        export RECCE_PROJECT_ID=7
        recce cloud list-sessions --organization 10 --project 9
    """
    from rich.console import Console
    from rich.table import Table

    console = Console()
    handle_debug_flag(**kwargs)

    try:
        api_token = prepare_api_token(**kwargs)
    except RecceConfigException:
        show_invalid_api_token_message()
        exit(1)

    organization = kwargs.get("organization")
    project = kwargs.get("project")

    # Validate required parameters
    if not organization:
        console.print("[[red]Error[/red]] Organization ID is required. Please provide it via:")
        console.print("  --organization <id> or set RECCE_ORGANIZATION_ID environment variable")
        exit(1)

    if not project:
        console.print("[[red]Error[/red]] Project ID is required. Please provide it via:")
        console.print("  --project <id> or set RECCE_PROJECT_ID environment variable")
        exit(1)

    try:
        from recce.util.recce_cloud import RecceCloud

        cloud = RecceCloud(api_token)
        sessions = cloud.list_sessions(organization, project)

        if not sessions:
            console.print(f"No sessions found in project {project}.")
            return

        table = Table(title=f"Sessions in Project {project}")
        table.add_column("ID", style="cyan")
        table.add_column("Name", style="green")
        table.add_column("Is Base", style="yellow")

        for session in sessions:
            is_base = "✓" if session.get("is_base", False) else ""
            table.add_row(session.get("id", ""), session.get("name", ""), is_base)

        console.print(table)

    except RecceCloudException as e:
        console.print(f"[[red]Error[/red]] {e}")
        exit(1)
    except Exception as e:
        console.print(f"[[red]Error[/red]] {e}")
        exit(1)


@cli.group("github", short_help="GitHub related commands", hidden=True)
def github(**kwargs):
    pass


@github.command(
    cls=TrackCommand, short_help="Download the artifacts from the GitHub repository based on the current Pull Request."
)
@click.option(
    "--github-token",
    help="The github token to use for accessing GitHub repo.",
    type=click.STRING,
    envvar="GITHUB_TOKEN",
)
@click.option(
    "--github-repo",
    help="The github repo to use for accessing GitHub repo.",
    type=click.STRING,
    envvar="GITHUB_REPOSITORY",
)
def artifact(**kwargs):
    from recce.github import recce_ci_artifact

    return recce_ci_artifact(**kwargs)


@cli.command(cls=TrackCommand)
@click.argument("state_file", type=click.Path(exists=True))
@click.option(
    "--api-token", help="The personal token generated by Recce Cloud.", type=click.STRING, envvar="RECCE_API_TOKEN"
)
def share(state_file, **kwargs):
    """
    Share the state file
    """
    from rich.console import Console

    console = Console()
    handle_debug_flag(**kwargs)
    cloud_options = None

    # read or input the api token
    try:
        api_token = prepare_api_token(interaction=True, **kwargs)
    except Abort:
        console.print("[yellow]Abort[/yellow]")
        exit(0)
    except RecceConfigException:
        show_invalid_api_token_message()
        exit(1)

    auth_options = {"api_token": api_token}

    # load local state
    state_loader = create_state_loader(
        review_mode=True, cloud_mode=False, state_file=state_file, cloud_options=cloud_options
    )

    if not state_loader.verify():
        error, hint = state_loader.error_and_hint
        console.print(f"[[red]Error[/red]] {error}")
        console.print(f"{hint}")
        exit(1)

    # check if state exists in cloud
    state_manager = RecceShareStateManager(auth_options)
    if not state_manager.verify():
        error, hint = state_manager.error_and_hint
        console.print(f"[[red]Error[/red]] {error}")
        console.print(f"{hint}")
        exit(1)

    # check if state exists in cloud
    state_file_name = os.path.basename(state_file)

    try:
        response = state_manager.share_state(state_file_name, state_loader.state)
        if response.get("status") == "error":
            console.print("[[red]Error[/red]] Failed to share the state.\n" f"Reason: {response.get('message')}")
        else:
            console.print(f"Shared Link: {response.get('share_url')}")
    except RecceCloudException as e:
        console.print(f"[[red]Error[/red]] {e}")
        console.print(f"Reason: {e.reason}")
        exit(1)


snapshot_id_option = click.option(
    "--snapshot-id",
    help="The snapshot ID to upload artifacts to cloud.",
    type=click.STRING,
    envvar=["RECCE_SNAPSHOT_ID", "RECCE_SESSION_ID"],
    required=True,
)

session_id_option = click.option(
    "--session-id",
    help="The session ID to upload artifacts to cloud.",
    type=click.STRING,
    envvar=["RECCE_SESSION_ID", "RECCE_SNAPSHOT_ID"],
    required=True,
)

target_path_option = click.option(
    "--target-path",
    help="dbt artifacts directory for your artifacts.",
    type=click.STRING,
    default="target",
    show_default=True,
)


@cli.command(cls=TrackCommand, hidden=True)
@add_options([session_id_option, target_path_option])
@add_options(recce_cloud_auth_options)
@add_options(recce_options)
def upload_session(**kwargs):
    """
    Upload target/manifest.json and target/catalog.json to the specific session ID

    Upload the dbt artifacts (manifest.json, catalog.json) to Recce Cloud for the given session ID.
    This allows you to associate artifacts with a specific session for later use.

    Examples:\n

    \b
    # Upload artifacts to a session ID
    recce upload-session --session-id <session-id>

    \b
    # Upload artifacts from custom target path to a session ID
    recce upload-session --session-id <session-id> --target-path my-target
    """
    from rich.console import Console

    console = Console()
    handle_debug_flag(**kwargs)

    # Initialize Recce Config
    RecceConfig(config_file=kwargs.get("config"))

    try:
        api_token = prepare_api_token(**kwargs)
    except RecceConfigException:
        show_invalid_api_token_message()
        exit(1)

    session_id = kwargs.get("session_id")
    target_path = kwargs.get("target_path")

    try:
        rc = upload_artifacts_to_session(
            target_path, session_id=session_id, token=api_token, debug=kwargs.get("debug", False)
        )
        console.rule("Uploaded Successfully")
        console.print(
            f'Uploaded dbt artifacts to Recce Cloud for session ID "{session_id}" from "{os.path.abspath(target_path)}"'
        )
    except Exception as e:
        console.rule("Failed to Upload Session", style="red")
        console.print(f"[[red]Error[/red]] Failed to upload the dbt artifacts to the session {session_id}.")
        console.print(f"Reason: {e}")
        rc = 1
    return rc


# Backward compatibility for `recce snapshot` command
@cli.command(
    cls=TrackCommand,
    hidden=True,
    deprecated=True,
    help="Upload target/manifest.json and target/catalog.json to the specific snapshot ID",
)
@add_options([snapshot_id_option, target_path_option])
@add_options(recce_cloud_auth_options)
@add_options(recce_options)
def snapshot(**kwargs):
    kwargs["session_id"] = kwargs.get("snapshot_id")
    return upload_session(**kwargs)


@cli.command(hidden=True, cls=TrackCommand)
@click.argument("state_file", required=True)
@click.option("--host", default="localhost", show_default=True, help="The host to bind to.")
@click.option("--port", default=8000, show_default=True, help="The port to bind to.", type=int)
@click.option("--lifetime", default=0, show_default=True, help="The lifetime of the server in seconds.", type=int)
@click.option("--share-url", help="The share URL triggers this instance.", type=click.STRING, envvar="RECCE_SHARE_URL")
@click.pass_context
def read_only(ctx, state_file=None, **kwargs):
    # Invoke `recce server --mode read-only <state_file> ...
    kwargs["mode"] = RecceServerMode.read_only
    ctx.invoke(server, state_file=state_file, **kwargs)


@cli.command(cls=TrackCommand)
@add_options(dbt_related_options)
@add_options(sqlmesh_related_options)
@add_options(recce_options)
@add_options(recce_dbt_artifact_dir_options)
@add_options(recce_cloud_options)
@add_options(recce_cloud_auth_options)
@add_options(recce_hidden_options)
def mcp_server(**kwargs):
    """
    [Experiment] Start the Recce MCP (Model Context Protocol) server

    The MCP server provides a stdio-based interface for AI assistants and tools
    to interact with Recce's data validation capabilities.

    Available tools:
    - get_lineage_diff: Get lineage differences between environments
    - row_count_diff: Compare row counts between environments
    - query: Execute SQL queries with dbt templating
    - query_diff: Compare query results between environments
    - profile_diff: Generate statistical profiles and compare

    Examples:\n

    \b
    # Start the MCP server
    recce mcp-server

    \b
    # Start with custom dbt configuration
    recce mcp-server --target prod --project-dir ./my_project
    """
    from rich.console import Console

    console = Console()
    try:
        # Import here to avoid import errors if mcp is not installed
        from recce.mcp_server import run_mcp_server
    except ImportError as e:
        console.print(f"[[red]Error[/red]] Failed to import MCP server: {e}")
        console.print(r"Please install the MCP package: pip install 'recce\[mcp]'")
        exit(1)

    # Initialize Recce Config
    RecceConfig(config_file=kwargs.get("config"))

    handle_debug_flag(**kwargs)
    patch_derived_args(kwargs)

    # Prepare API token
    try:
        api_token = prepare_api_token(**kwargs)
        kwargs["api_token"] = api_token
    except RecceConfigException:
        show_invalid_api_token_message()
        exit(1)

    # Create state loader using shared function (if cloud mode is enabled)
    is_cloud = kwargs.get("cloud", False)
    if is_cloud:
        state_loader = create_state_loader_by_args(None, **kwargs)
        kwargs["state_loader"] = state_loader

    try:
        console.print("Starting Recce MCP Server...")
        console.print("Available tools: get_lineage_diff, row_count_diff, query, query_diff, profile_diff")

        # Run the async server
        asyncio.run(run_mcp_server(**kwargs))
    except Exception as e:
        console.print(f"[[red]Error[/red]] Failed to start MCP server: {e}")
        if kwargs.get("debug"):
            import traceback

            traceback.print_exc()
        exit(1)


if __name__ == "__main__":
    cli()



import json
import typing as t
from decimal import Decimal
from enum import Enum

if t.TYPE_CHECKING:
    import agate
    import pandas
from pydantic import BaseModel, Field


class DataFrameColumnType(Enum):
    NUMBER = "number"
    INTEGER = "integer"
    TEXT = "text"
    BOOLEAN = "boolean"
    DATE = "date"
    DATETIME = "datetime"
    TIMEDELTA = "timedelta"
    UNKNOWN = "unknown"


class DataFrameColumn(BaseModel):
    name: str
    type: DataFrameColumnType


class DataFrame(BaseModel):
    columns: t.List[DataFrameColumn]
    data: t.List[tuple]
    limit: t.Optional[int] = Field(None, description="Limit the number of rows returned")
    more: t.Optional[bool] = Field(None, description="Whether there are more rows to fetch")

    @staticmethod
    def from_agate(table: "agate.Table", limit: t.Optional[int] = None, more: t.Optional[bool] = None):
        from recce.adapter.dbt_adapter import dbt_version

        if dbt_version < "v1.8":
            import dbt.clients.agate_helper as agate_helper
        else:
            import dbt_common.clients.agate_helper as agate_helper

        import agate

        columns = []

        for col_name, col_type in zip(table.column_names, table.column_types):

            has_integer = hasattr(agate_helper, "Integer")

            if isinstance(col_type, agate.Number):
                col_type = DataFrameColumnType.NUMBER
            elif isinstance(col_type, agate.Text):
                col_type = DataFrameColumnType.TEXT
            elif isinstance(col_type, agate.Boolean):
                col_type = DataFrameColumnType.BOOLEAN
            elif isinstance(col_type, agate.Date):
                col_type = DataFrameColumnType.DATE
            elif isinstance(col_type, agate.DateTime):
                col_type = DataFrameColumnType.DATETIME
            elif isinstance(col_type, agate.TimeDelta):
                col_type = DataFrameColumnType.TIMEDELTA
            elif has_integer and isinstance(col_type, agate_helper.Integer):
                col_type = DataFrameColumnType.INTEGER
            else:
                col_type = DataFrameColumnType.UNKNOWN
            columns.append(DataFrameColumn(name=col_name, type=col_type))

        def _row_values(row):
            # If the value is Decimal, check if it's finite. If not, convert it to float(xxx) (GitHub issue #476)
            return tuple([float(v) if isinstance(v, Decimal) and not v.is_finite() else v for v in row.values()])

        data = [_row_values(row) for row in table.rows]
        df = DataFrame(
            columns=columns,
            data=data,
            limit=limit,
            more=more,
        )
        return df

    @staticmethod
    def from_pandas(pandas_df: "pandas.DataFrame", limit: t.Optional[int] = None, more: t.Optional[bool] = None):
        columns = []
        for column in pandas_df.columns:
            dtype = pandas_df[column].dtype
            if dtype == "int64":
                col_type = DataFrameColumnType.INTEGER
            elif dtype == "float64":
                col_type = DataFrameColumnType.NUMBER
            elif dtype == "object":
                col_type = DataFrameColumnType.TEXT
            elif dtype == "bool":
                col_type = DataFrameColumnType.BOOLEAN
            else:
                col_type = DataFrameColumnType.UNKNOWN
            columns.append(DataFrameColumn(name=column, type=col_type))

        s = pandas_df.to_json(orient="values")
        data = json.loads(s)

        df = DataFrame(
            columns=columns,
            data=data,
            limit=limit,
            more=more,
        )
        return df



"""
Test that CLI can be imported even when mcp is not available.
"""

import sys
from unittest.mock import patch


def test_cli_can_be_imported_without_mcp():
    """Test that recce.cli can be imported even if mcp package is not available"""
    # This test verifies that the CLI module doesn't fail to import
    # when mcp is not installed, since mcp is an optional dependency
    from recce import cli

    assert cli is not None
    assert hasattr(cli, "cli")
    assert hasattr(cli, "mcp_server")


def test_mcp_server_command_fails_gracefully_without_mcp():
    """Test that mcp-server command shows helpful error when mcp is not available"""
    # Mock sys.modules to simulate mcp not being installed
    with patch.dict(sys.modules, {"mcp": None, "mcp.server": None, "mcp.server.stdio": None, "mcp.types": None}):
        # Remove mcp_server from modules to force reimport
        if "recce.mcp_server" in sys.modules:
            del sys.modules["recce.mcp_server"]

        from recce.cli import mcp_server

        # The function should exist
        assert mcp_server is not None

        # When called, it should handle ImportError gracefully
        # (We can't easily test the actual execution without more mocking,
        # but we've verified the function exists and can be imported)


def test_cli_command_exists():
    """Test that both server and mcp-server commands are registered"""
    from recce.cli import cli

    # Check that both commands exist
    commands = {cmd.name for cmd in cli.commands.values()}
    assert "server" in commands
    assert "mcp_server" in commands or "mcp-server" in commands



import time
from dataclasses import dataclass
from typing import Dict, List, Optional, Tuple

import sqlglot.expressions as exp
from sqlglot import Dialect, parse_one
from sqlglot.errors import OptimizeError, SqlglotError
from sqlglot.optimizer import Scope, traverse_scope
from sqlglot.optimizer.qualify import qualify

from recce.exceptions import RecceException
from recce.models.types import CllColumn, CllColumnDep

CllResult = Tuple[
    List[CllColumnDep],  # Model to column dependencies
    Dict[str, CllColumn],  # Column to column dependencies
]


@dataclass
class CLLPerformanceTracking:
    lineage_start = None
    lineage_elapsed = None
    column_lineage_start = None
    column_lineage_elapsed = None

    total_nodes = None
    sqlglot_error_nodes = 0
    other_error_nodes = 0

    def start_lineage(self):
        self.lineage_start = time.perf_counter_ns()

    def end_lineage(self):
        if self.lineage_start is None:
            return
        self.lineage_elapsed = (time.perf_counter_ns() - self.lineage_start) / 1000000

    def start_column_lineage(self):
        self.column_lineage_start = time.perf_counter_ns()

    def end_column_lineage(self):
        if self.column_lineage_start is None:
            return
        self.column_lineage_elapsed = (time.perf_counter_ns() - self.column_lineage_start) / 1000000

    def set_total_nodes(self, total_nodes):
        self.total_nodes = total_nodes

    def increment_sqlglot_error_nodes(self):
        self.sqlglot_error_nodes += 1

    def increment_other_error_nodes(self):
        self.other_error_nodes += 1

    def to_dict(self):
        return {
            "lineage_elapsed_ms": self.lineage_elapsed,
            "column_lineage_elapsed_ms": self.column_lineage_elapsed,
            "total_nodes": self.total_nodes,
            "sqlglot_error_nodes": self.sqlglot_error_nodes,
            "other_error_nodes": self.other_error_nodes,
        }

    def reset(self):
        self.lineage_start = None
        self.lineage_elapsed = None
        self.column_lineage_start = None
        self.column_lineage_elapsed = None

        self.total_nodes = None
        self.sqlglot_error_nodes = 0
        self.other_error_nodes = 0


def _cll_column(proj, table_alias_map) -> CllColumn:
    # given an expression, return the columns depends on
    # [{node: table, column: column}, ...]
    type = "source"
    depends_on: List[CllColumnDep] = []

    # instance of Column
    if isinstance(proj, exp.Alias):
        # 'select a as b'
        # 'select CURRENT_TIMESTAMP() as create_at'
        root = proj.this

    for expression in root.walk(bfs=False):
        if isinstance(expression, exp.Column):
            column = expression
            alias = column.table

            if alias is None:
                table = next(iter(table_alias_map.values()))
            else:
                table = table_alias_map.get(alias, alias)
            depends_on.append(CllColumnDep(table, column.name))
            if type == "source":
                type = "passthrough"
        elif isinstance(expression, (exp.Paren, exp.Identifier)):
            pass
        else:
            type = "derived"

    depends_on = _dedeup_depends_on(depends_on)

    if len(depends_on) == 0:
        type = "source"

    if isinstance(proj, exp.Alias):
        alias = proj
        if type == "passthrough" and depends_on[0].column != alias.alias_or_name:
            type = "renamed"

    return CllColumn(type=type, depends_on=depends_on)


def _dedeup_depends_on(depends_on: List[CllColumnDep]) -> List[CllColumnDep]:
    # deduplicate the depends_on list
    dedup_set = set()
    dedup_list = []
    for col_dep in depends_on:
        node_col = col_dep.node + "." + col_dep.column
        if node_col not in dedup_set:
            dedup_list.append(col_dep)
            dedup_set.add(node_col)
    return dedup_list


def _cll_set_scope(scope: Scope, scope_cll_map: dict[Scope, CllResult]) -> CllResult:
    # model-to-column
    m2c: List[CllColumnDep] = []
    # column-to-column
    c2c_map: Dict[str, CllColumn] = {}

    for union_scope in scope.union_scopes:
        sub_scope_result = scope_cll_map.get(union_scope)
        if sub_scope_result is None:
            raise RecceException(f"Scope {union_scope} not found in scope_cll_map")
        sub_m2c, sub_c2c_map = sub_scope_result

        for k, v in sub_c2c_map.items():
            if k not in c2c_map:
                c2c_map[k] = v
            else:
                c2c_map[k].depends_on.extend(v.depends_on)
                c2c_map[k].transformation_type = "derived"

        m2c.extend(sub_m2c)
    return m2c, c2c_map


def _cll_select_scope(scope: Scope, scope_cll_map: dict[Scope, CllResult]) -> CllResult:
    assert scope.expression.key == "select"

    # model-to-column
    m2c: List[CllColumnDep] = []
    # column-to-column
    c2c_map: Dict[str, CllColumn] = {}

    table_alias_map = {t.alias_or_name: t.name for t in scope.tables}
    select = scope.expression

    def source_column_dependency(ref_column: exp.Column) -> Optional[CllColumn]:
        column_name = ref_column.name
        table_name = ref_column.table if ref_column.table != "" else next(iter(table_alias_map.values()))
        source = scope.sources.get(table_name, None)  # transformation_type: exp.Table | Scope
        if isinstance(source, Scope):
            ref_cll_result = scope_cll_map.get(source)
            if ref_cll_result is None:
                return None
            _, sub_c2c_map = ref_cll_result
            return sub_c2c_map.get(column_name)
        elif isinstance(source, exp.Table):
            return CllColumn(
                name=column_name,
                transformation_type="passthrough",
                depends_on=[CllColumnDep(node=source.name, column=column_name)],
            )
        else:
            return None

    def subquery_cll(subquery: exp.Subquery) -> Optional[CllResult]:
        select = subquery.find(exp.Select)
        if select is None:
            return None

        matched_scope = None
        for sub_scope in scope.subquery_scopes:
            if sub_scope.expression == select:
                matched_scope = sub_scope
                break
        if matched_scope is None:
            return None

        return scope_cll_map.get(matched_scope)

    for proj in scope.expression.selects:
        transformation_type = "source"
        column_depends_on: List[CllColumnDep] = []
        root = proj.this if isinstance(proj, exp.Alias) else proj
        for expression in root.walk(bfs=False):
            if isinstance(expression, exp.Column):
                ref_column_dependency = source_column_dependency(expression)
                if ref_column_dependency is not None:
                    column_depends_on.extend(ref_column_dependency.depends_on)
                    if ref_column_dependency.transformation_type == "derived":
                        transformation_type = "derived"
                    elif ref_column_dependency.transformation_type == "renamed":
                        if transformation_type == "source" or transformation_type == "passthrough":
                            transformation_type = "renamed"
                    elif ref_column_dependency.transformation_type == "passthrough":
                        if transformation_type == "source":
                            transformation_type = "passthrough"
                else:
                    column_depends_on.append(CllColumnDep(expression.table, expression.name))
                    if transformation_type == "source":
                        transformation_type = "passthrough"

            elif isinstance(expression, (exp.Paren, exp.Identifier)):
                pass
            else:
                transformation_type = "derived"

        column_depends_on = _dedeup_depends_on(column_depends_on)

        if len(column_depends_on) == 0 and transformation_type != "source":
            transformation_type = "source"

        if isinstance(proj, exp.Alias):
            alias = proj
            if transformation_type == "passthrough" and column_depends_on[0].column != alias.alias_or_name:
                transformation_type = "renamed"

        c2c_map[proj.alias_or_name] = CllColumn(
            name=proj.alias_or_name, transformation_type=transformation_type, depends_on=column_depends_on
        )

    def selected_column_dependency(ref_column: exp.Column) -> Optional[CllColumn]:
        column_name = ref_column.name
        return c2c_map.get(column_name)

    # joins clause: Reference the source columns
    if select.args.get("joins"):
        joins = select.args.get("joins")
        for join in joins:
            if isinstance(join, exp.Join):
                for ref_column in join.find_all(exp.Column):
                    if source_column_dependency(ref_column) is not None:
                        m2c.extend(source_column_dependency(ref_column).depends_on)

    # where clauses: Reference the source columns
    if select.args.get("where"):
        where = select.args.get("where")
        if isinstance(where, exp.Where):
            for ref_column in where.find_all(exp.Column):
                if source_column_dependency(ref_column) is not None:
                    m2c.extend(source_column_dependency(ref_column).depends_on)
            for subquery in where.find_all(exp.Subquery):
                sub_cll = subquery_cll(subquery)
                if sub_cll is not None:
                    sub_m2c, sub_c2c_map = sub_cll
                    m2c.extend(sub_m2c)
                    for sub_c in sub_c2c_map.values():
                        m2c.extend(sub_c.depends_on)

    # group by clause: Reference the source columns, column index
    if select.args.get("group"):
        group = select.args.get("group")
        if isinstance(group, exp.Group):
            for ref_column in group.find_all(exp.Column):
                if source_column_dependency(ref_column) is not None:
                    m2c.extend(source_column_dependency(ref_column).depends_on)

    # having clause: Reference the source columns, selected columns
    if select.args.get("having"):
        having = select.args.get("having")
        if isinstance(having, exp.Having):
            for ref_column in having.find_all(exp.Column):
                if source_column_dependency(ref_column) is not None:
                    m2c.extend(source_column_dependency(ref_column).depends_on)
                elif selected_column_dependency(ref_column) is not None:
                    m2c.extend(selected_column_dependency(ref_column).depends_on)
            for subquery in having.find_all(exp.Subquery):
                sub_cll = subquery_cll(subquery)
                if sub_cll is not None:
                    sub_m2c, sub_c2c_map = sub_cll
                    m2c.extend(sub_m2c)
                    for sub_c in sub_c2c_map.values():
                        m2c.extend(sub_c.depends_on)

    # order by clause: Reference the source columns, selected columns, column index
    if select.args.get("order"):
        order = select.args.get("order")
        if isinstance(order, exp.Order):
            for ref_column in order.find_all(exp.Column):
                if source_column_dependency(ref_column) is not None:
                    m2c.extend(source_column_dependency(ref_column).depends_on)
                elif selected_column_dependency(ref_column) is not None:
                    m2c.extend(selected_column_dependency(ref_column).depends_on)

    for source in scope.sources.values():
        scope_cll_result = scope_cll_map.get(source)
        if scope_cll_result is None:
            continue
        sub_m2c, _ = scope_cll_result
        m2c.extend(sub_m2c)

    m2c = _dedeup_depends_on(m2c)

    return m2c, c2c_map


def cll(sql, schema=None, dialect=None) -> CllResult:
    # given a sql, return the cll for the sql
    # {
    #     'depends_on': [{'node': 'model_id', 'column': 'column'}],
    #     'columns': {
    #         'column1': {
    #             'type': 'derived',
    #             'depends_on': [{'node': 'model_id', 'column': 'column'}],
    #         }
    #     }
    # }

    dialect = Dialect.get(dialect) if dialect is not None else None

    try:
        expression = parse_one(sql, dialect=dialect)
    except SqlglotError as e:
        raise RecceException(f"Failed to parse SQL: {str(e)}")

    try:
        expression = qualify(expression, schema=schema, dialect=dialect)
    except OptimizeError as e:
        raise RecceException(f"Failed to optimize SQL: {str(e)}")
    except SqlglotError as e:
        raise RecceException(f"Failed to qualify SQL: {str(e)}")

    result = None
    scope_cll_map = {}
    for scope in traverse_scope(expression):
        scope_type = scope.expression.key
        if scope_type == "union" or scope_type == "intersect" or scope_type == "except":
            result = _cll_set_scope(scope, scope_cll_map)
        elif scope_type == "select":
            result = _cll_select_scope(scope, scope_cll_map)
        else:
            continue

        scope_cll_map[scope] = result

    if result is None:
        raise RecceException("Failed to extract CLL from SQL")
    return result



import pytest

from recce.tasks import QueryDiffTask, QueryTask


def test_query_diff_in_client(dbt_test_helper):
    csv_data_curr = """
        customer_id,name,age
        1,Alice,30
        2,Bob,25
        3,Charlie,35
        """

    csv_data_base = """
        customer_id,name,age
        1,Alice,35
        2,Bob,25
        3,Charlie,35
        """

    dbt_test_helper.create_model("customers", csv_data_base, csv_data_curr)
    params = dict(sql_template='select * from {{ ref("customers") }}')
    task = QueryTask(params)
    run_result = task.execute()
    assert len(run_result.current.data) == 3


def test_query_diff_in_client(dbt_test_helper):
    csv_data_curr = """
        customer_id,name,age
        1,Alice,30
        2,Bob,25
        3,Charlie,35
        """

    csv_data_base = """
        customer_id,name,age
        1,Alice,35
        2,Bob,25
        3,Charlie,35
        """

    dbt_test_helper.create_model("customers", csv_data_base, csv_data_curr)
    params = dict(sql_template='select * from {{ ref("customers") }}')
    task = QueryDiffTask(params)
    run_result = task.execute()
    assert len(run_result.base.data) == 3
    assert len(run_result.current.data) == 3

    params = dict(
        base_sql_template='select * from {{ ref("customers") }} where customer_id <= 2',
        sql_template='select * from {{ ref("customers") }}',
    )
    task = QueryDiffTask(params)
    run_result = task.execute()
    assert len(run_result.base.data) == 2
    assert len(run_result.current.data) == 3


def test_query_diff_in_warehouse(dbt_test_helper):
    csv_data_curr = """
        customer_id,name,age
        1,Alice,30
        2,Bob,25
        3,Charlie,35
        """

    csv_data_base = """
        customer_id,name,age
        1,Alice,35
        2,Bob,25
        3,Charlie,35
        """

    dbt_test_helper.create_model("customers", csv_data_base, csv_data_curr)
    params = dict(sql_template='select * from {{ ref("customers") }}', primary_keys=["customer_id"])
    task = QueryDiffTask(params)
    run_result = task.execute()
    assert len(run_result.diff.data) == 2

    params = dict(
        base_sql_template='select * from {{ ref("customers") }} where customer_id == 1',
        sql_template='select * from {{ ref("customers") }}',
        primary_keys=["customer_id"],
    )
    task = QueryDiffTask(params)
    run_result = task.execute()
    assert len(run_result.diff.data) == 4


def test_validator():
    from recce.tasks.query import QueryCheckValidator, QueryDiffCheckValidator

    def validate(params: dict = {}, view_options: dict = {}):
        QueryCheckValidator().validate(
            {
                "name": "test",
                "type": "query",
                "params": params,
                "view_options": view_options,
            }
        )

    def validate_diff(params: dict = {}, view_options: dict = {}):
        QueryDiffCheckValidator().validate(
            {
                "name": "test",
                "type": "query_diff",
                "params": params,
                "view_options": view_options,
            }
        )

    # query
    validate({"sql_template": "select * from abc"})

    # diff in client
    validate_diff({"sql_template": "select * from abc"})
    validate_diff(
        {
            "sql_template": "select * from abc",
            "base_sql_template": "select * from abc",
        }
    )

    # diff in warehouse
    validate_diff(
        {
            "primary_keys": ["customer_id"],
            "sql_template": "select * from abc",
        }
    )
    validate_diff(
        {"sql_template": "select * from abc", "base_sql_template": "select * from abc", "primary_keys": ["customer_id"]}
    )
    validate_diff(
        {"sql_template": "select * from abc", "base_sql_template": "select * from abc", "primary_keys": ["customer_id"]}
    )

    # invalid
    with pytest.raises(ValueError):
        validate()

    with pytest.raises(ValueError):
        validate_diff()

    with pytest.raises(ValueError):
        validate_diff({"sql_template": 123, "primary_keys": "xyz"})

    with pytest.raises(ValueError):
        validate_diff({"sql_template": "s", "primary_keys": "xyz"})



import json
import unittest
from unittest.mock import MagicMock, patch

from recce_cloud.api.client import RecceCloudClient, RecceCloudException


class RecceCloudClientTests(unittest.TestCase):
    def setUp(self):
        """Set up test fixtures."""
        self.api_token = "rct-test-token-123"
        self.session_id = "session-123"
        self.org_id = "org-456"
        self.project_id = "project-789"

    def test_init_with_api_token(self):
        """Test client initialization with Recce API token."""
        client = RecceCloudClient(self.api_token)
        self.assertEqual(client.token, self.api_token)
        self.assertIn("/api/v2", client.base_url_v2)

    def test_init_with_none_token_raises_error(self):
        """Test client initialization with None token raises ValueError."""
        with self.assertRaises(ValueError) as context:
            RecceCloudClient(None)
        self.assertIn("Token cannot be None", str(context.exception))

    @patch("recce_cloud.api.client.requests.request")
    def test_get_session_success(self, mock_request):
        """Test successful get_session call."""
        client = RecceCloudClient(self.api_token)

        # Mock successful response
        mock_response = MagicMock()
        mock_response.status_code = 200
        mock_response.json.return_value = {
            "success": True,
            "session": {
                "id": self.session_id,
                "org_id": self.org_id,
                "project_id": self.project_id,
            },
        }
        mock_request.return_value = mock_response

        result = client.get_session(self.session_id)

        self.assertEqual(result["id"], self.session_id)
        self.assertEqual(result["org_id"], self.org_id)
        self.assertEqual(result["project_id"], self.project_id)

        # Verify request was made correctly
        mock_request.assert_called_once()
        call_args = mock_request.call_args
        self.assertEqual(call_args[0][0], "GET")
        self.assertIn(self.session_id, call_args[0][1])
        self.assertEqual(call_args[1]["headers"]["Authorization"], f"Bearer {self.api_token}")

    @patch("recce_cloud.api.client.requests.request")
    def test_get_session_not_found(self, mock_request):
        """Test get_session with 404 response."""
        client = RecceCloudClient(self.api_token)

        # Mock 404 response
        mock_response = MagicMock()
        mock_response.status_code = 404
        mock_response.text = "Session not found"
        mock_request.return_value = mock_response

        with self.assertRaises(RecceCloudException) as context:
            client.get_session(self.session_id)

        self.assertEqual(context.exception.status_code, 404)
        self.assertIn("Failed to get session", str(context.exception))

    @patch("recce_cloud.api.client.requests.request")
    def test_get_session_forbidden(self, mock_request):
        """Test get_session with 403 response."""
        client = RecceCloudClient(self.api_token)

        # Mock 403 response
        mock_response = MagicMock()
        mock_response.status_code = 403
        mock_response.json.return_value = {"detail": "Access denied"}
        mock_request.return_value = mock_response

        result = client.get_session(self.session_id)

        self.assertEqual(result["status"], "error")
        self.assertEqual(result["message"], "Access denied")

    @patch("recce_cloud.api.client.requests.request")
    def test_get_session_api_error(self, mock_request):
        """Test get_session with API returning success=False."""
        client = RecceCloudClient(self.api_token)

        # Mock response with success=False
        mock_response = MagicMock()
        mock_response.status_code = 200
        mock_response.json.return_value = {"success": False, "message": "Invalid session"}
        mock_request.return_value = mock_response

        with self.assertRaises(RecceCloudException) as context:
            client.get_session(self.session_id)

        self.assertIn("Invalid session", context.exception.reason)

    @patch("recce_cloud.api.client.requests.request")
    def test_get_upload_urls_success(self, mock_request):
        """Test successful get_upload_urls_by_session_id call."""
        client = RecceCloudClient(self.api_token)

        # Mock successful response
        mock_response = MagicMock()
        mock_response.status_code = 200
        mock_response.json.return_value = {
            "presigned_urls": {
                "manifest_url": "https://s3.amazonaws.com/bucket/manifest.json?token=abc",
                "catalog_url": "https://s3.amazonaws.com/bucket/catalog.json?token=def",
            }
        }
        mock_request.return_value = mock_response

        result = client.get_upload_urls_by_session_id(self.org_id, self.project_id, self.session_id)

        self.assertIn("manifest_url", result)
        self.assertIn("catalog_url", result)
        self.assertIn("s3.amazonaws.com", result["manifest_url"])

        # Verify request was made correctly
        mock_request.assert_called_once()
        call_args = mock_request.call_args
        self.assertEqual(call_args[0][0], "GET")
        self.assertIn(self.org_id, call_args[0][1])
        self.assertIn(self.project_id, call_args[0][1])
        self.assertIn(self.session_id, call_args[0][1])
        self.assertIn("upload-url", call_args[0][1])

    @patch("recce_cloud.api.client.requests.request")
    def test_get_upload_urls_no_presigned_urls(self, mock_request):
        """Test get_upload_urls_by_session_id with no presigned URLs."""
        client = RecceCloudClient(self.api_token)

        # Mock response with null presigned_urls
        mock_response = MagicMock()
        mock_response.status_code = 200
        mock_response.json.return_value = {"presigned_urls": None}
        mock_request.return_value = mock_response

        with self.assertRaises(RecceCloudException) as context:
            client.get_upload_urls_by_session_id(self.org_id, self.project_id, self.session_id)

        self.assertEqual(context.exception.status_code, 404)
        self.assertIn("No presigned URLs", str(context.exception))

    @patch("recce_cloud.api.client.requests.request")
    def test_get_upload_urls_failure(self, mock_request):
        """Test get_upload_urls_by_session_id with API failure."""
        client = RecceCloudClient(self.api_token)

        # Mock error response
        mock_response = MagicMock()
        mock_response.status_code = 500
        mock_response.text = "Internal server error"
        mock_request.return_value = mock_response

        with self.assertRaises(RecceCloudException) as context:
            client.get_upload_urls_by_session_id(self.org_id, self.project_id, self.session_id)

        self.assertEqual(context.exception.status_code, 500)

    @patch("recce_cloud.api.client.requests.request")
    def test_update_session_success(self, mock_request):
        """Test successful update_session call."""
        client = RecceCloudClient(self.api_token)
        adapter_type = "postgres"

        # Mock successful response
        mock_response = MagicMock()
        mock_response.status_code = 200
        mock_response.json.return_value = {"success": True, "session": {"adapter_type": adapter_type}}
        mock_request.return_value = mock_response

        result = client.update_session(self.org_id, self.project_id, self.session_id, adapter_type)

        self.assertTrue(result["success"])
        self.assertEqual(result["session"]["adapter_type"], adapter_type)

        # Verify request was made correctly
        mock_request.assert_called_once()
        call_args = mock_request.call_args
        self.assertEqual(call_args[0][0], "PATCH")
        self.assertIn(self.org_id, call_args[0][1])
        self.assertIn(self.project_id, call_args[0][1])
        self.assertIn(self.session_id, call_args[0][1])
        self.assertEqual(call_args[1]["json"]["adapter_type"], adapter_type)

    @patch("recce_cloud.api.client.requests.request")
    def test_update_session_forbidden(self, mock_request):
        """Test update_session with 403 response."""
        client = RecceCloudClient(self.api_token)

        # Mock 403 response
        mock_response = MagicMock()
        mock_response.status_code = 403
        mock_response.json.return_value = {"detail": "Insufficient permissions"}
        mock_request.return_value = mock_response

        result = client.update_session(self.org_id, self.project_id, self.session_id, "postgres")

        self.assertEqual(result["status"], "error")
        self.assertEqual(result["message"], "Insufficient permissions")

    @patch("recce_cloud.api.client.requests.request")
    def test_update_session_failure(self, mock_request):
        """Test update_session with API failure."""
        client = RecceCloudClient(self.api_token)

        # Mock error response
        mock_response = MagicMock()
        mock_response.status_code = 400
        mock_response.text = "Bad request"
        mock_request.return_value = mock_response

        with self.assertRaises(RecceCloudException) as context:
            client.update_session(self.org_id, self.project_id, self.session_id, "invalid_adapter")

        self.assertEqual(context.exception.status_code, 400)

    def test_recce_cloud_exception_with_json_detail(self):
        """Test RecceCloudException parses JSON detail."""
        json_reason = json.dumps({"detail": "Invalid session ID"})
        exception = RecceCloudException(message="Test error", reason=json_reason, status_code=400)

        self.assertEqual(exception.status_code, 400)
        self.assertEqual(exception.reason, "Invalid session ID")
        self.assertIn("Test error", str(exception))

    def test_recce_cloud_exception_with_plain_text(self):
        """Test RecceCloudException with plain text reason."""
        plain_reason = "Connection timeout"
        exception = RecceCloudException(message="Test error", reason=plain_reason, status_code=500)

        self.assertEqual(exception.status_code, 500)
        self.assertEqual(exception.reason, plain_reason)

    @patch.dict("os.environ", {"RECCE_INSTANCE_ENV": "docker"})
    @patch("recce_cloud.api.client.requests.request")
    def test_docker_internal_url_replacement(self, mock_request):
        """Test localhost URL is replaced with docker internal URL."""
        client = RecceCloudClient(self.api_token)

        # Mock response with localhost URL
        mock_response = MagicMock()
        mock_response.status_code = 200
        mock_response.json.return_value = {
            "presigned_urls": {
                "manifest_url": "http://localhost:8000/manifest.json",
                "catalog_url": "http://localhost:8000/catalog.json",
            }
        }
        mock_request.return_value = mock_response

        result = client.get_upload_urls_by_session_id(self.org_id, self.project_id, self.session_id)

        # URLs should be replaced with docker internal
        self.assertIn("host.docker.internal", result["manifest_url"])
        self.assertIn("host.docker.internal", result["catalog_url"])
        self.assertNotIn("localhost", result["manifest_url"])


if __name__ == "__main__":
    unittest.main()



import pytest

from recce.tasks import TopKDiffTask
from recce.tasks.top_k import TopKDiffCheckValidator


def test_top_k(dbt_test_helper):
    csv_data_curr = """
        customer_id,name,age
        1,Alice,30
        2,Bob,25
        3,Charlie,35
        4,Bob,35
        """

    csv_data_base = """
        customer_id,name,age
        1,Alice,35
        2,Bob,25
        3,Charlie,35
        4,,35
        """

    dbt_test_helper.create_model("customers", csv_data_base, csv_data_curr)

    params = dict(
        model="customers",
        column_name="name",
        k=50,
    )

    task = TopKDiffTask(params)
    run_result = task.execute()

    # {
    #     'values': ['Bob', 'Alice', 'Charlie'],
    #     'counts': [2, 1, 1],
    #     'valids': 4,
    #     'total': 4
    # }
    assert run_result["current"]["values"][0] == "Bob"
    assert run_result["current"]["counts"][0] == 2
    assert run_result["current"]["valids"] == 4
    assert run_result["current"]["total"] == 4

    # {
    #     'values': ['Bob', 'Alice', 'Charlie'],
    #     'counts': [1, 1, 1],
    #     'valids': 3,
    #     'total': 4
    # }
    assert run_result["base"]["counts"][0] == 1
    assert run_result["base"]["valids"] == 3
    assert run_result["base"]["total"] == 4


def test_validator():
    def validate(params: dict = {}, view_options: dict = {}):
        TopKDiffCheckValidator().validate(
            {
                "name": "test",
                "type": "top_k_diff",
                "params": params,
                "view_options": view_options,
            }
        )

    validate(
        {
            "model": "customers",
            "column_name": "name",
            "k": 50,
        }
    )

    with pytest.raises(ValueError):
        validate({})



from recce.adapter.dbt_adapter import DbtAdapter, dbt_supported_registry


def test_dbt_adapter_support_tasks(dbt_test_helper):
    adapter: DbtAdapter = dbt_test_helper.context.adapter

    # Test dbt task support
    support_tasks = adapter.support_tasks()

    for task_type in dbt_supported_registry:
        task = task_type.value
        assert task in support_tasks
        assert support_tasks[task] is True


def test_dbt_adapter_support_tasks_without_required_dbt_package(dbt_test_helper):
    adapter: DbtAdapter = dbt_test_helper.context.adapter
    # Mock the macros in manifest to simulate no required dbt package installed
    adapter.manifest.macros = {}

    support_tasks = adapter.support_tasks()

    for task_type in dbt_supported_registry:
        task = task_type.value
        assert task in support_tasks



import click
from rich.console import Console

from recce import event
from recce.event import get_recce_api_token, update_recce_api_token
from recce.exceptions import RecceConfigException
from recce.util.recce_cloud import (
    RECCE_CLOUD_BASE_URL,
    RecceCloud,
)

console = Console()


def show_invalid_api_token_message():
    """
    Show the message when the API token is invalid.
    """
    console.print("[[red]Error[/red]] Invalid Recce Cloud API token.")
    console.print("Please associate with your Recce Cloud account by the following command 'recce connect-to-cloud'.")
    console.print(
        "For more information, please visit: https://docs.reccehq.com/recce-cloud/share-recce-session-securely/#configure-recce-cloud-association-manually"
    )


def prepare_api_token(
    interaction=False,
    **kwargs,
):
    """
    Prepare the API token for the request.
    """
    # Verify the API token for Recce Cloud Share Link
    api_token = get_recce_api_token()
    new_api_token = kwargs.get("api_token")
    if new_api_token is not None and new_api_token.startswith("rct-"):
        # Task Token
        valid = RecceCloud(new_api_token).verify_token()
        if not valid:
            raise RecceConfigException("Invalid Recce Cloud Task token")
        api_token = new_api_token
    elif api_token != new_api_token and new_api_token is not None:
        # Handle the API token provided by option `--api-token`
        valid = RecceCloud(new_api_token).verify_token()
        if not valid:
            raise RecceConfigException("Invalid Recce Cloud API token")
        event.log_connected_to_cloud()
        api_token = new_api_token
        update_recce_api_token(api_token)
        console.print(
            "[[green]Success[/green]] User profile has been updated to include the Recce Cloud API Token. "
            "You no longer need to append --api-token to the recce command"
        )
    elif api_token:
        # Verify the API token from the user profile
        valid = RecceCloud(api_token).verify_token()
        if not valid:
            console.print("[[yellow]Warning[/yellow]] Invalid Recce Cloud API token. Skipping the share link.")
            api_token = None
        if valid:
            event.log_connected_to_cloud()
    else:
        # No api_token provided
        if interaction:
            console.print(
                "An API token is required for this feature. This can be obtained in your user account settings.\n"
                f"{RECCE_CLOUD_BASE_URL}/settings#tokens\n"
                "Your API token can be added to '~/.recce/profile.yml' for more convenient sharing."
            )
            api_token = click.prompt("Your Recce API token", type=str, hide_input=True, show_default=False)
            valid = RecceCloud(api_token).verify_token()
            if not valid:
                raise RecceConfigException("Invalid Recce Cloud API token")
            update_recce_api_token(api_token)
            console.print(
                "[[green]Success[/green]] User profile has been updated to include the Recce Cloud API Token. "
                "You no longer need to append --api-token to the recce command"
            )

    return api_token



import os
import textwrap
import uuid
from datetime import datetime
from io import StringIO

from dbt.contracts.graph.nodes import (
    ModelNode,
    SeedNode,
    SnapshotNode,
    SourceDefinition,
)
from dbt.contracts.results import (
    CatalogArtifact,
    CatalogTable,
    ColumnMetadata,
    TableMetadata,
)

from recce.adapter.dbt_adapter import DbtAdapter, as_manifest, load_manifest
from recce.core import RecceContext


class DbtTestHelper:

    def __init__(self):
        schema_prefix = "schema_" + uuid.uuid4().hex
        self.base_schema = f"{schema_prefix}_base"
        self.curr_schema = f"{schema_prefix}_curr"

        current_dir = os.path.dirname(os.path.abspath(__file__))
        project_dir = os.path.join(current_dir, "test_proj")
        profiles_dir = project_dir
        manifest_path = os.path.join(project_dir, "manifest.json")

        dbt_adapter = DbtAdapter.load(
            no_artifacts=True,
            project_dir=project_dir,
            profiles_dir=profiles_dir,
        )

        context = RecceContext()
        context.adapter_type = "dbt"
        context.adapter = dbt_adapter
        context.schema_prefix = schema_prefix
        self.adapter = dbt_adapter
        self.context = context
        curr_writable_manifest = load_manifest(manifest_path)
        base_writable_manifest = load_manifest(manifest_path)

        self.curr_manifest = as_manifest(curr_writable_manifest)
        self.base_manifest = as_manifest(base_writable_manifest)
        now = datetime.now()
        self.curr_catalog = CatalogArtifact.from_results(
            generated_at=now,
            nodes={},
            sources={},
            compile_results=None,
            errors=None,
        )
        self.base_catalog = CatalogArtifact.from_results(
            generated_at=now,
            nodes={},
            sources={},
            compile_results=None,
            errors=None,
        )
        self.context = context

        self.adapter.execute(f"CREATE schema IF NOT EXISTS {self.base_schema}")
        self.adapter.execute(f"CREATE schema IF NOT EXISTS {self.curr_schema}")
        self.adapter.set_artifacts(
            base_writable_manifest,
            curr_writable_manifest,
            self.curr_manifest,
            self.base_manifest,
            self.base_catalog,
            self.curr_catalog,
        )

    def create_model(
        self,
        model_name,
        base_csv=None,
        curr_csv=None,
        base_sql=None,
        curr_sql=None,
        depends_on=None,
        disabled=False,
        unique_id=None,
        resource_type="model",
        package_name="recce_test",
        base_columns: dict[str, str] = None,
        curr_columns: dict[str, str] = None,
        patch_func=None,
    ):
        # unique_id = f"model.{package_name}.{model_name}"
        unique_id = unique_id if unique_id else model_name
        if depends_on is None:
            depends_on = []

        def _add_model_to_manifest(base):
            if base:
                schema = self.base_schema
                manifest = self.base_manifest
                catalog = self.base_catalog
                csv = base_csv
                sql = base_sql
                columns = base_columns
            else:
                schema = self.curr_schema
                manifest = self.curr_manifest
                catalog = self.curr_catalog
                csv = curr_csv
                sql = curr_sql
                columns = curr_columns

            if csv:
                dbt_adapter = self.adapter
                csv = textwrap.dedent(csv)
                with dbt_adapter.connection_named("create model"):
                    import pandas as pd

                    df = pd.read_csv(StringIO(csv))  # noqa: F841
                    # DuckDB disables python_scan_all_frames by default after 1.1,
                    # ref: https://github.com/duckdb/duckdb/pull/13896
                    dbt_adapter.execute("SET python_scan_all_frames=true")
                    dbt_adapter.execute(f"CREATE TABLE {schema}.{model_name} AS SELECT * FROM df")
            raw_code = sql if sql else csv

            if columns:
                index = 1
                table = CatalogTable(TableMetadata(type="BASE TABLE", schema=schema, name=model_name), {}, {})
                catalog.nodes[unique_id] = table
                for column, column_type in columns.items():
                    col_data = ColumnMetadata(type=column_type, index=index, name=column)
                    catalog.nodes[unique_id].columns[column] = col_data
                    index = index + 1

            node_dict = {
                "resource_type": resource_type,
                "name": model_name,
                "package_name": package_name,
                "path": "",
                "original_file_path": "",
                "unique_id": unique_id,
                "fqn": [
                    package_name,
                    model_name,
                ],
                "schema": schema,
                "alias": model_name,
                "checksum": {
                    "name": "sha256",
                    "checksum": hash(raw_code),
                },
                "raw_code": raw_code,
                "config": {
                    "materialized": "table",
                    "tags": ["test_tag"],
                },
                "tags": ["test_tag"],
                "depends_on": {"nodes": depends_on},
            }
            if patch_func:
                patch_func(node_dict)

            if resource_type == "snapshot":
                node = SnapshotNode.from_dict(node_dict)
            elif resource_type == "seed":
                node = SeedNode.from_dict(node_dict)
            else:
                node = ModelNode.from_dict(node_dict)

            if disabled:
                manifest.add_disabled_nofile(node)
            else:
                manifest.add_node_nofile(node)
            return node

        if base_csv or base_sql:
            _add_model_to_manifest(True)

        if curr_csv or curr_sql:
            _add_model_to_manifest(False)

        self.adapter.set_artifacts(
            self.base_manifest.writable_manifest(),
            self.curr_manifest.writable_manifest(),
            self.curr_manifest,
            self.base_manifest,
            self.base_catalog,
            self.curr_catalog,
        )

    def remove_model(self, model_name):
        dbt_adapter = self.adapter
        with dbt_adapter.connection_named("cleanup"):
            dbt_adapter.execute(f"DROP TABLE IF EXISTS {self.base_schema}.{model_name}")
            dbt_adapter.execute(f"DROP TABLE IF EXISTS {self.curr_schema}.{model_name} ")

    def create_source(
        self,
        source_name,
        table_name,
        base_csv=None,
        curr_csv=None,
        unique_id=None,
        package_name="recce_test",
        base_columns: dict[str, str] = None,
        curr_columns: dict[str, str] = None,
        patch_func=None,
    ):
        unique_id = unique_id if unique_id else f"source.{package_name}.{source_name}.{table_name}"

        def _add_source_to_manifest(base):
            if base:
                schema = self.base_schema
                manifest = self.base_manifest
                catalog = self.base_catalog
                csv = base_csv
                columns = base_columns
            else:
                schema = self.curr_schema
                manifest = self.curr_manifest
                catalog = self.curr_catalog
                csv = curr_csv
                columns = curr_columns

            if csv:
                dbt_adapter = self.adapter
                csv = textwrap.dedent(csv)
                with dbt_adapter.connection_named("create source"):
                    import pandas as pd

                    df = pd.read_csv(StringIO(csv))  # noqa: F841
                    # DuckDB disables python_scan_all_frames by default after 1.1,
                    # ref: https://github.com/duckdb/duckdb/pull/13896
                    dbt_adapter.execute("SET python_scan_all_frames=true")
                    dbt_adapter.execute(f"CREATE TABLE {schema}.{table_name} AS SELECT * FROM df")

            if columns:
                index = 1
                table = CatalogTable(TableMetadata(type="BASE TABLE", schema=schema, name=table_name), {}, {})
                catalog.sources[unique_id] = table
                for column, column_type in columns.items():
                    col_data = ColumnMetadata(type=column_type, index=index, name=column)
                    catalog.sources[unique_id].columns[column] = col_data
                    index = index + 1

            node_dict = {
                "resource_type": "source",
                "source_name": source_name,
                "name": table_name,
                "identifier": table_name,
                "source_description": "test source",
                "loader": "",
                "package_name": package_name,
                "path": "",
                "original_file_path": "",
                "unique_id": unique_id,
                "fqn": [
                    package_name,
                    source_name,
                    table_name,
                ],
                "schema": schema,
            }
            if patch_func:
                patch_func(node_dict)

            source = SourceDefinition.from_dict(node_dict)
            manifest.sources[unique_id] = source
            return source

        if base_csv:
            _add_source_to_manifest(True)

        if curr_csv:
            _add_source_to_manifest(False)

        self.adapter.set_artifacts(
            self.base_manifest.writable_manifest(),
            self.curr_manifest.writable_manifest(),
            self.curr_manifest,
            self.base_manifest,
            self.base_catalog,
            self.curr_catalog,
        )

    def cleanup(self):
        dbt_adapter = self.adapter
        with dbt_adapter.connection_named("cleanup"):
            dbt_adapter.execute(f"DROP SCHEMA IF EXISTS {self.base_schema} CASCADE")
            dbt_adapter.execute(f"DROP SCHEMA IF EXISTS {self.curr_schema} CASCADE")

    def create_snapshot(self, sanpshot_name, base_csv, curr_csv, depends_on=[]):
        self.create_model(sanpshot_name, base_csv, curr_csv, depends_on=depends_on, resource_type="snapshot")



# Configuration file for recce
from rich.console import Console

from recce import yaml
from recce.exceptions import RecceConfigException
from recce.util import SingletonMeta

RECCE_CONFIG_FILE = "recce.yml"
RECCE_PRESET_CHECK_COMMENT = """Preset Checks
Please see https://docs.datarecce.io/features/preset-checks/
"""
RECCE_ERROR_LOG_FILE = "recce_error.log"
console = Console()


class RecceConfig(metaclass=SingletonMeta):
    def __init__(self, config_file=RECCE_CONFIG_FILE):
        self.config_file = config_file
        self.config = None
        self.load()

    def load(self):
        try:
            with open(self.config_file, "r", encoding="utf-8") as f:
                config = yaml.safe_load(f)
                self.config = config if config else {}
            self._verify_preset_checks()
        except FileNotFoundError:
            console.print(f"[[orange3]NOTICE[/orange3]] Generate default Recce config file at '{self.config_file}'")
            self.config = self.generate_template()
            self.save()

    def _verify_preset_checks(self):
        from recce.tasks.core import CheckValidator

        if not self.config.get("checks"):
            return

        for check in self.config["checks"]:
            try:
                check_type = check.get("type")
                if check_type is None:
                    raise ValueError(f'Check type is required for check "{check}"')
                if check_type == "lineage_diff":
                    from recce.tasks.lineage import LineageDiffCheckValidator

                    validator = LineageDiffCheckValidator()
                elif check_type == "schema_diff":
                    from recce.tasks.schema import SchemaDiffCheckValidator

                    validator = SchemaDiffCheckValidator()
                elif check_type == "row_count_diff":
                    from recce.tasks.rowcount import RowCountDiffCheckValidator

                    validator = RowCountDiffCheckValidator()
                elif check_type == "query":
                    from recce.tasks.query import QueryCheckValidator

                    validator = QueryCheckValidator()
                elif check_type == "query_diff":
                    from recce.tasks.query import QueryDiffCheckValidator

                    validator = QueryDiffCheckValidator()
                elif check_type == "value_diff" or check_type == "value_diff_detail":
                    from recce.tasks.valuediff import ValueDiffCheckValidator

                    validator = ValueDiffCheckValidator()
                elif check_type == "profile_diff":
                    from recce.tasks.profile import ProfileCheckValidator

                    validator = ProfileCheckValidator()
                elif check_type == "top_k_diff":
                    from recce.tasks.top_k import TopKDiffCheckValidator

                    validator = TopKDiffCheckValidator()
                elif check_type == "histogram_diff":
                    from recce.tasks.histogram import HistogramDiffCheckValidator

                    validator = HistogramDiffCheckValidator()
                else:
                    validator = CheckValidator()
                validator.validate(check)
            except Exception as e:
                import json

                raise RecceConfigException(
                    f"Load preset checks failed from '{self.config_file}'\n{json.dumps(check, indent=2)}", cause=e
                )

    def generate_template(self):
        data = yaml.CommentedMap(checks=yaml.CommentedSeq())
        data.yaml_set_comment_before_after_key("checks", before=RECCE_PRESET_CHECK_COMMENT)
        # Define default preset checks
        default_checks = [
            yaml.CommentedMap(
                name="Row count diff",
                description="Check the row count diff for all table models.",
                type="row_count_diff",
                params={"select": "state:modified,config.materialized:table"},
            ),
            yaml.CommentedMap(
                name="Schema diff",
                description="Check the schema diff for all nodes.",
                type="schema_diff",
            ),
        ]

        for check in default_checks:
            data["checks"].append(check)

        return data

    def get(self, key, default=None):
        return self.config.get(key, default)

    def set(self, key, value):
        self.config[key] = value

    def save(self):
        with open(RECCE_CONFIG_FILE, "w", encoding="utf-8") as f:
            yaml.dump(self.config, f)

    def __str__(self):
        return yaml.dump(self.config)

    def __repr__(self):
        return self.__str__()



import unittest
from unittest.mock import Mock, patch

from click.testing import CliRunner

from recce.cli import list_organizations, list_projects, list_sessions
from recce.exceptions import RecceConfigException
from recce.util.recce_cloud import RecceCloudException


class TestCloudListingCLI(unittest.TestCase):
    """Test cases for the cloud listing CLI commands."""

    def setUp(self):
        """Set up test fixtures."""
        self.runner = CliRunner()

    @patch("recce.cli.prepare_api_token")
    @patch("recce.util.recce_cloud.RecceCloud")
    def test_list_organizations_success(self, mock_recce_cloud, mock_prepare_token):
        """Test successful list-organizations command."""
        # Setup mocks
        mock_prepare_token.return_value = "test-token"
        mock_cloud_instance = Mock()
        mock_cloud_instance.list_organizations.return_value = [
            {"id": 1, "name": "org1", "display_name": "Organization 1"},
            {"id": 2, "name": "org2", "display_name": "Organization 2"},
        ]
        mock_recce_cloud.return_value = mock_cloud_instance

        # Test command
        result = self.runner.invoke(list_organizations, [])

        # Assertions
        self.assertEqual(result.exit_code, 0)
        self.assertIn("Organizations", result.output)
        self.assertIn("org1", result.output)
        self.assertIn("Organization 1", result.output)
        self.assertIn("org2", result.output)
        self.assertIn("Organization 2", result.output)
        mock_cloud_instance.list_organizations.assert_called_once()

    @patch("recce.cli.prepare_api_token")
    @patch("recce.util.recce_cloud.RecceCloud")
    def test_list_organizations_empty(self, mock_recce_cloud, mock_prepare_token):
        """Test list-organizations command with no organizations."""
        # Setup mocks
        mock_prepare_token.return_value = "test-token"
        mock_cloud_instance = Mock()
        mock_cloud_instance.list_organizations.return_value = []
        mock_recce_cloud.return_value = mock_cloud_instance

        # Test command
        result = self.runner.invoke(list_organizations, [])

        # Assertions
        self.assertEqual(result.exit_code, 0)
        self.assertIn("No organizations found", result.output)

    @patch("recce.cli.prepare_api_token")
    def test_list_organizations_invalid_token(self, mock_prepare_token):
        """Test list-organizations command with invalid token."""
        # Setup mock to raise exception
        mock_prepare_token.side_effect = RecceConfigException("Invalid token")

        # Test command
        result = self.runner.invoke(list_organizations, [])

        # Assertions
        self.assertEqual(result.exit_code, 1)

    @patch("recce.cli.prepare_api_token")
    @patch("recce.util.recce_cloud.RecceCloud")
    def test_list_projects_with_cli_arg(self, mock_recce_cloud, mock_prepare_token):
        """Test list-projects command with CLI argument."""
        # Setup mocks
        mock_prepare_token.return_value = "test-token"
        mock_cloud_instance = Mock()
        mock_cloud_instance.list_projects.return_value = [
            {"id": 1, "name": "project1", "display_name": "Project 1"},
            {"id": 2, "name": "project2", "display_name": "Project 2"},
        ]
        mock_recce_cloud.return_value = mock_cloud_instance

        # Test command with CLI argument
        result = self.runner.invoke(list_projects, ["--organization", "8"])

        # Assertions
        self.assertEqual(result.exit_code, 0)
        self.assertIn("Projects in Organization 8", result.output)
        self.assertIn("project1", result.output)
        self.assertIn("Project 1", result.output)
        mock_cloud_instance.list_projects.assert_called_once_with("8")

    @patch("recce.cli.prepare_api_token")
    @patch("recce.util.recce_cloud.RecceCloud")
    def test_list_projects_with_env_var(self, mock_recce_cloud, mock_prepare_token):
        """Test list-projects command with environment variable."""
        # Setup mocks
        mock_prepare_token.return_value = "test-token"
        mock_cloud_instance = Mock()
        mock_cloud_instance.list_projects.return_value = [{"id": 1, "name": "project1", "display_name": "Project 1"}]
        mock_recce_cloud.return_value = mock_cloud_instance

        # Test command with environment variable
        result = self.runner.invoke(list_projects, [], env={"RECCE_ORGANIZATION_ID": "8"})

        # Assertions
        self.assertEqual(result.exit_code, 0)
        self.assertIn("Projects in Organization 8", result.output)
        self.assertIn("project1", result.output)
        mock_cloud_instance.list_projects.assert_called_once_with("8")

    @patch("recce.cli.prepare_api_token")
    @patch("recce.util.recce_cloud.RecceCloud")
    def test_list_projects_cli_overrides_env(self, mock_recce_cloud, mock_prepare_token):
        """Test list-projects command where CLI argument overrides environment variable."""
        # Setup mocks
        mock_prepare_token.return_value = "test-token"
        mock_cloud_instance = Mock()
        mock_cloud_instance.list_projects.return_value = [{"id": 1, "name": "project1", "display_name": "Project 1"}]
        mock_recce_cloud.return_value = mock_cloud_instance

        # Test command with both env var and CLI arg (CLI should win)
        result = self.runner.invoke(list_projects, ["--organization", "10"], env={"RECCE_ORGANIZATION_ID": "8"})

        # Assertions
        self.assertEqual(result.exit_code, 0)
        self.assertIn("Projects in Organization 10", result.output)
        # Verify CLI argument (10) was used, not env var (8)
        mock_cloud_instance.list_projects.assert_called_once_with("10")

    @patch("recce.cli.prepare_api_token")
    def test_list_projects_missing_organization(self, mock_prepare_token):
        """Test list-projects command with missing organization ID."""
        # Setup mocks
        mock_prepare_token.return_value = "test-token"

        # Test command without organization ID
        result = self.runner.invoke(list_projects, [])

        # Assertions
        self.assertEqual(result.exit_code, 1)
        self.assertIn("Organization ID is required", result.output)
        self.assertIn("--organization", result.output)
        self.assertIn("RECCE_ORGANIZATION_ID", result.output)

    @patch("recce.cli.prepare_api_token")
    @patch("recce.util.recce_cloud.RecceCloud")
    def test_list_projects_empty(self, mock_recce_cloud, mock_prepare_token):
        """Test list-projects command with no projects."""
        # Setup mocks
        mock_prepare_token.return_value = "test-token"
        mock_cloud_instance = Mock()
        mock_cloud_instance.list_projects.return_value = []
        mock_recce_cloud.return_value = mock_cloud_instance

        # Test command
        result = self.runner.invoke(list_projects, ["--organization", "8"])

        # Assertions
        self.assertEqual(result.exit_code, 0)
        self.assertIn("No projects found in organization 8", result.output)

    @patch("recce.cli.prepare_api_token")
    @patch("recce.util.recce_cloud.RecceCloud")
    def test_list_sessions_with_cli_args(self, mock_recce_cloud, mock_prepare_token):
        """Test list-sessions command with CLI arguments."""
        # Setup mocks
        mock_prepare_token.return_value = "test-token"
        mock_cloud_instance = Mock()
        mock_cloud_instance.list_sessions.return_value = [
            {"id": "session1", "name": "PR-123", "is_base": False},
            {"id": "session2", "name": "Base Session", "is_base": True},
        ]
        mock_recce_cloud.return_value = mock_cloud_instance

        # Test command with CLI arguments
        result = self.runner.invoke(list_sessions, ["--organization", "8", "--project", "7"])

        # Assertions
        self.assertEqual(result.exit_code, 0)
        self.assertIn("Sessions in Project 7", result.output)
        self.assertIn("PR-123", result.output)
        self.assertIn("Base Session", result.output)
        self.assertIn("✓", result.output)  # Base session marker
        mock_cloud_instance.list_sessions.assert_called_once_with("8", "7")

    @patch("recce.cli.prepare_api_token")
    @patch("recce.util.recce_cloud.RecceCloud")
    def test_list_sessions_with_env_vars(self, mock_recce_cloud, mock_prepare_token):
        """Test list-sessions command with environment variables."""
        # Setup mocks
        mock_prepare_token.return_value = "test-token"
        mock_cloud_instance = Mock()
        mock_cloud_instance.list_sessions.return_value = [{"id": "session1", "name": "Session 1", "is_base": False}]
        mock_recce_cloud.return_value = mock_cloud_instance

        # Test command with environment variables
        result = self.runner.invoke(list_sessions, [], env={"RECCE_ORGANIZATION_ID": "8", "RECCE_PROJECT_ID": "7"})

        # Assertions
        self.assertEqual(result.exit_code, 0)
        self.assertIn("Sessions in Project 7", result.output)
        self.assertIn("Session 1", result.output)
        mock_cloud_instance.list_sessions.assert_called_once_with("8", "7")

    @patch("recce.cli.prepare_api_token")
    @patch("recce.util.recce_cloud.RecceCloud")
    def test_list_sessions_mixed_env_and_cli(self, mock_recce_cloud, mock_prepare_token):
        """Test list-sessions command with mixed environment variables and CLI args."""
        # Setup mocks
        mock_prepare_token.return_value = "test-token"
        mock_cloud_instance = Mock()
        mock_cloud_instance.list_sessions.return_value = [{"id": "session1", "name": "Session 1", "is_base": False}]
        mock_recce_cloud.return_value = mock_cloud_instance

        # Test command with env var for org and CLI arg for project
        result = self.runner.invoke(list_sessions, ["--project", "9"], env={"RECCE_ORGANIZATION_ID": "8"})

        # Assertions
        self.assertEqual(result.exit_code, 0)
        self.assertIn("Sessions in Project 9", result.output)
        # Verify it used env var for org (8) and CLI arg for project (9)
        mock_cloud_instance.list_sessions.assert_called_once_with("8", "9")

    @patch("recce.cli.prepare_api_token")
    def test_list_sessions_missing_organization(self, mock_prepare_token):
        """Test list-sessions command with missing organization ID."""
        # Setup mocks
        mock_prepare_token.return_value = "test-token"

        # Test command without organization ID
        result = self.runner.invoke(list_sessions, ["--project", "7"])

        # Assertions
        self.assertEqual(result.exit_code, 1)
        self.assertIn("Organization ID is required", result.output)
        self.assertIn("--organization", result.output)
        self.assertIn("RECCE_ORGANIZATION_ID", result.output)

    @patch("recce.cli.prepare_api_token")
    def test_list_sessions_missing_project(self, mock_prepare_token):
        """Test list-sessions command with missing project ID."""
        # Setup mocks
        mock_prepare_token.return_value = "test-token"

        # Test command without project ID
        result = self.runner.invoke(list_sessions, ["--organization", "8"])

        # Assertions
        self.assertEqual(result.exit_code, 1)
        self.assertIn("Project ID is required", result.output)
        self.assertIn("--project", result.output)
        self.assertIn("RECCE_PROJECT_ID", result.output)

    @patch("recce.cli.prepare_api_token")
    @patch("recce.util.recce_cloud.RecceCloud")
    def test_list_sessions_empty(self, mock_recce_cloud, mock_prepare_token):
        """Test list-sessions command with no sessions."""
        # Setup mocks
        mock_prepare_token.return_value = "test-token"
        mock_cloud_instance = Mock()
        mock_cloud_instance.list_sessions.return_value = []
        mock_recce_cloud.return_value = mock_cloud_instance

        # Test command
        result = self.runner.invoke(list_sessions, ["--organization", "8", "--project", "7"])

        # Assertions
        self.assertEqual(result.exit_code, 0)
        self.assertIn("No sessions found in project 7", result.output)

    @patch("recce.cli.prepare_api_token")
    @patch("recce.util.recce_cloud.RecceCloud")
    def test_list_sessions_api_error(self, mock_recce_cloud, mock_prepare_token):
        """Test list-sessions command with API error."""
        # Setup mocks
        mock_prepare_token.return_value = "test-token"
        mock_cloud_instance = Mock()
        mock_cloud_instance.list_sessions.side_effect = RecceCloudException("Access denied", "Forbidden", 403)
        mock_recce_cloud.return_value = mock_cloud_instance

        # Test command
        result = self.runner.invoke(list_sessions, ["--organization", "8", "--project", "7"])

        # Assertions
        self.assertEqual(result.exit_code, 1)
        self.assertIn("Error", result.output)

    @patch("recce.cli.prepare_api_token")
    @patch("recce.util.recce_cloud.RecceCloud")
    def test_sessions_base_session_display(self, mock_recce_cloud, mock_prepare_token):
        """Test that base sessions are properly marked with checkmark."""
        # Setup mocks
        mock_prepare_token.return_value = "test-token"
        mock_cloud_instance = Mock()
        mock_cloud_instance.list_sessions.return_value = [
            {"id": "session1", "name": "Regular Session", "is_base": False},
            {"id": "session2", "name": "Base Session", "is_base": True},
            {"id": "session3", "name": "Another Regular", "is_base": False},
        ]
        mock_recce_cloud.return_value = mock_cloud_instance

        # Test command
        result = self.runner.invoke(list_sessions, ["--organization", "8", "--project", "7"])

        # Assertions
        self.assertEqual(result.exit_code, 0)
        output_lines = result.output.split("\n")

        # Find the base session line and verify it has the checkmark
        base_session_line = None
        for line in output_lines:
            if "Base Session" in line:
                base_session_line = line
                break

        self.assertIsNotNone(base_session_line)
        self.assertIn("✓", base_session_line)


if __name__ == "__main__":
    unittest.main()



import logging
import threading
import time
from abc import ABC, abstractmethod
from typing import Dict, Literal, Optional, Tuple, Union, final

from recce.exceptions import RecceException
from recce.pull_request import fetch_pr_metadata

from ..util.io import SupportedFileTypes, file_io_factory
from .const import RECCE_API_TOKEN_MISSING
from .state import RecceState

logger = logging.getLogger("uvicorn")


class RecceStateLoader(ABC):
    def __init__(
        self,
        review_mode: bool = False,
        cloud_mode: bool = False,
        state_file: Optional[str] = None,
        cloud_options: Optional[Dict[str, str]] = None,
        initial_state: Optional[RecceState] = None,
    ):
        self.review_mode = review_mode
        self.cloud_mode = cloud_mode
        self.state_file = state_file
        self.cloud_options = cloud_options or {}
        self.error_message = None
        self.hint_message = None
        self.state: RecceState | None = initial_state
        self.state_lock = threading.Lock()
        self.state_etag = None
        self.pr_info = None
        self.catalog: Literal["github", "preview", "snapshot"] = "github"
        self.share_id = None
        self.session_id = None

        if self.cloud_mode:
            if self.cloud_options.get("github_token"):
                self.catalog = "github"
                self.pr_info = fetch_pr_metadata(
                    cloud=self.cloud_mode, github_token=self.cloud_options.get("github_token")
                )
                if self.pr_info.id is None:
                    raise RecceException("Cannot get the pull request information from GitHub.")
            elif self.cloud_options.get("api_token"):
                if self.cloud_options.get("session_id"):
                    self.catalog = "session"
                    self.session_id = self.cloud_options.get("session_id")
                else:
                    self.catalog = "preview"
                    self.share_id = self.cloud_options.get("share_id")
            else:
                raise RecceException(RECCE_API_TOKEN_MISSING.error_message)

    @property
    def token(self):
        return self.cloud_options.get("github_token") or self.cloud_options.get("api_token")

    @abstractmethod
    def verify(self) -> bool:
        """
        Verify the state loader configuration.
        Returns:
            bool: True if the configuration is valid, False otherwise.
        """
        raise NotImplementedError("Subclasses must implement this method.")

    @property
    def error_and_hint(self) -> (Union[str, None], Union[str, None]):
        return self.error_message, self.hint_message

    def update(self, state: RecceState):
        self.state = state

    @final
    def load(self, refresh=False) -> RecceState:
        if self.state is not None and refresh is False:
            return self.state
        self.state_lock.acquire()
        try:
            self.state, self.state_etag = self._load_state()
        finally:
            self.state_lock.release()
        return self.state

    @abstractmethod
    def _load_state(self) -> Tuple[RecceState, str]:
        """
        Load the state from the specified source (file or cloud).
        Returns:
            RecceState: The loaded state object.
            str: The etag of the state file (if applicable).
        """
        raise NotImplementedError("Subclasses must implement this method.")

    def save_as(self, state_file: str, state: RecceState = None):
        if self.cloud_mode:
            raise Exception("Cannot save the state to Recce Cloud.")

        self.state_file = state_file
        self.export(state)

    @final
    def export(self, state: RecceState = None) -> Union[str, None]:
        if state is not None:
            self.update(state)

        start_time = time.time()
        self.state_lock.acquire()
        try:
            message, state_etag = self._export_state()
            self.state_etag = state_etag
            end_time = time.time()
            elapsed_time = end_time - start_time
        finally:
            self.state_lock.release()
        logger.info(f"Store state completed in {elapsed_time:.2f} seconds")
        return message

    @abstractmethod
    def _export_state(self) -> Tuple[Union[str, None], str]:
        """
        Export the current Recce state to a file or cloud storage.
        Returns:
            str: A message indicating the result of the export operation.
            str: The etag of the exported state file (if applicable).
        """
        raise NotImplementedError("Subclasses must implement this method.")

    def _export_state_to_file(self, file_path: str, file_type: SupportedFileTypes = SupportedFileTypes.FILE) -> str:
        """
        Store the state to a file. Store happens when terminating the server or run instance.
        """

        json_data = self.state.to_json()
        io = file_io_factory(file_type)

        io.write(file_path, json_data)
        return f"The state file is stored at '{file_path}'"

    def refresh(self):
        new_state = self.load(refresh=True)
        return new_state

    def check_conflict(self) -> bool:
        return False

    def info(self) -> dict:
        if self.state is None:
            self.error_message = "No state is loaded."
            return None

        state_info = {
            "mode": "cloud" if self.cloud_mode else "local",
            "source": None,
        }
        if self.cloud_mode:
            state_info["source"] = "Recce Cloud"
            state_info["pull_request"] = self.pr_info
        else:
            state_info["source"] = self.state_file
        return state_info

    @abstractmethod
    def purge(self) -> bool:
        """
        Purge the state file or cloud storage.
        Returns:
            bool: True if the purge was successful, False otherwise.
        """
        raise NotImplementedError("Subclasses must implement this method.")



import logging
import os
from base64 import b64encode
from hashlib import md5, sha256
from typing import Dict, Optional, Tuple, Union
from urllib.parse import urlencode

from recce.exceptions import RecceException
from recce.pull_request import PullRequestInfo, fetch_pr_metadata
from recce.util.io import SupportedFileTypes, file_io_factory
from recce.util.recce_cloud import PresignedUrlMethod, RecceCloud, RecceCloudException

from ..event import get_recce_api_token
from ..models import CheckDAO
from .const import (
    RECCE_API_TOKEN_MISSING,
    RECCE_CLOUD_PASSWORD_MISSING,
    RECCE_CLOUD_TOKEN_MISSING,
    RECCE_STATE_COMPRESSED_FILE,
)
from .state import RecceState
from .state_loader import RecceStateLoader

logger = logging.getLogger("uvicorn")


def s3_sse_c_headers(password: str) -> Dict[str, str]:
    hashed_password = sha256()
    md5_hash = md5()
    hashed_password.update(password.encode())
    md5_hash.update(hashed_password.digest())
    encoded_passwd = b64encode(hashed_password.digest()).decode("utf-8")
    encoded_md5 = b64encode(md5_hash.digest()).decode("utf-8")
    return {
        "x-amz-server-side-encryption-customer-algorithm": "AES256",
        "x-amz-server-side-encryption-customer-key": encoded_passwd,
        "x-amz-server-side-encryption-customer-key-MD5": encoded_md5,
    }


class CloudStateLoader(RecceStateLoader):
    def __init__(
        self,
        review_mode: bool = False,
        cloud_options: Optional[Dict[str, str]] = None,
        initial_state: Optional[RecceState] = None,
    ):
        super().__init__(
            cloud_mode=True,
            review_mode=review_mode,
            cloud_options=cloud_options,
            initial_state=initial_state,
        )
        self.recce_cloud = RecceCloud(token=self.token)

    def verify(self) -> bool:
        if self.catalog == "github":
            if self.cloud_options.get("github_token") is None:
                self.error_message = RECCE_CLOUD_TOKEN_MISSING.error_message
                self.hint_message = RECCE_CLOUD_TOKEN_MISSING.hint_message
                return False
            if not self.cloud_options.get("host"):
                if self.cloud_options.get("password") is None:
                    self.error_message = RECCE_CLOUD_PASSWORD_MISSING.error_message
                    self.hint_message = RECCE_CLOUD_PASSWORD_MISSING.hint_message
                    return False
        elif self.catalog == "preview":
            if self.cloud_options.get("api_token") is None:
                self.error_message = RECCE_API_TOKEN_MISSING.error_message
                self.hint_message = RECCE_API_TOKEN_MISSING.hint_message
                return False
            if self.cloud_options.get("share_id") is None:
                self.error_message = "No share ID is provided for the preview catalog."
                self.hint_message = (
                    'Please provide a share URL in the command argument with option "--share-url <share-url>"'
                )
                return False
        elif self.catalog == "session":
            if self.cloud_options.get("api_token") is None:
                self.error_message = RECCE_API_TOKEN_MISSING.error_message
                self.hint_message = RECCE_API_TOKEN_MISSING.hint_message
                return False
            if self.cloud_options.get("session_id") is None:
                self.error_message = "No session ID is provided for the session catalog."
                self.hint_message = (
                    'Please provide a session ID in the command argument with option "--session-id <session-id>"'
                )
                return False
        return True

    def purge(self) -> bool:
        rc, err_msg = RecceCloudStateManager(self.cloud_options).purge_cloud_state()
        if err_msg:
            self.error_message = err_msg
        return rc

    def _load_state(self) -> Tuple[RecceState, str]:
        """
        Load the state from Recce Cloud based on catalog type.

        Returns:
            RecceState: The state object.
            str: The etag of the state file (only used for GitHub).
        """
        if self.catalog == "github":
            return self._load_state_from_github()
        elif self.catalog == "preview":
            return self._load_state_from_preview()
        elif self.catalog == "session":
            return self._load_state_from_session(), None
        else:
            raise RecceException(f"Unsupported catalog type: {self.catalog}")

    def _load_state_from_github(self) -> Tuple[RecceState, str]:
        """Load state from GitHub PR with etag checking."""
        if (self.pr_info is None) or (self.pr_info.id is None) or (self.pr_info.repository is None):
            raise RecceException("Cannot get the pull request information from GitHub.")

        logger.debug("Fetching GitHub state from Recce Cloud...")

        # Check metadata and etag for GitHub only
        metadata = self._get_metadata_from_recce_cloud()
        state_etag = metadata.get("etag") if metadata else None

        # Return cached state if etag matches
        if self.state_etag and state_etag == self.state_etag:
            return self.state, self.state_etag

        # Download state from GitHub
        presigned_url = self.recce_cloud.get_presigned_url_by_github_repo(
            method=PresignedUrlMethod.DOWNLOAD,
            pr_id=self.pr_info.id,
            repository=self.pr_info.repository,
            artifact_name=RECCE_STATE_COMPRESSED_FILE,
        )

        password = self.cloud_options.get("password")
        if password is None:
            raise RecceException(RECCE_CLOUD_PASSWORD_MISSING.error_message)

        headers = s3_sse_c_headers(password)
        loaded_state = self._download_state_from_url(presigned_url, SupportedFileTypes.GZIP, headers)

        # Handle the case where download returns None (404 error)
        if loaded_state is None:
            return None, state_etag

        return loaded_state, state_etag

    def _load_state_from_preview(self) -> Tuple[RecceState, None]:
        """Load state from preview share (no etag checking needed)."""
        if self.share_id is None:
            raise RecceException("Cannot load the share state from Recce Cloud. No share ID is provided.")

        logger.debug("Fetching preview state from Recce Cloud...")

        # Download state from preview share
        presigned_url = self.recce_cloud.get_presigned_url_by_share_id(
            method=PresignedUrlMethod.DOWNLOAD, share_id=self.share_id
        )

        loaded_state = self._download_state_from_url(presigned_url, SupportedFileTypes.FILE)

        # Handle the case where download returns None (404 error)
        if loaded_state is None:
            return None, None

        return loaded_state, None

    def _get_metadata_from_recce_cloud(self) -> Union[dict, None]:
        return self.recce_cloud.get_artifact_metadata(pr_info=self.pr_info) if self.pr_info else None

    def _download_state_from_url(
        self, presigned_url: str, file_type: SupportedFileTypes, headers: dict = None
    ) -> RecceState:
        """Download state file from presigned URL and convert to RecceState."""
        import tempfile

        import requests

        with tempfile.NamedTemporaryFile() as tmp:
            response = requests.get(presigned_url, headers=headers)

            if response.status_code == 404:
                self.error_message = "The state file is not found in Recce Cloud."
                return None
            elif response.status_code != 200:
                self.error_message = response.text
                error_msg = f"{response.status_code} Failed to download the state file from Recce Cloud."
                if headers:  # GitHub case with password
                    error_msg += " The password could be wrong."
                raise RecceException(error_msg)

            with open(tmp.name, "wb") as f:
                f.write(response.content)

            return RecceState.from_file(tmp.name, file_type=file_type)

    def _load_state_from_session(self) -> RecceState:
        """
        Load state from session by:
        1. Get session info
        2. Download artifacts for both base and current sessions
        3. Download recce_state if available, otherwise create empty state with artifacts
        """
        if self.session_id is None:
            raise RecceException("Cannot load the session state from Recce Cloud. No session ID is provided.")

        # 1. Get session information
        logger.debug(f"Getting session {self.session_id}")
        session = self.recce_cloud.get_session(self.session_id)

        pr_url = session.get("pr_link")
        org_id = session.get("org_id")
        project_id = session.get("project_id")

        if not org_id or not project_id:
            raise RecceException(f"Session {self.session_id} does not belong to a valid organization or project.")

        # 2. Download manifests and catalogs for both session
        logger.debug(f"Downloading current session artifacts for {self.session_id}")
        current_artifacts = self._download_session_artifacts(self.recce_cloud, org_id, project_id, self.session_id)

        logger.debug(f"Downloading base session artifacts for project {project_id}")
        base_artifacts = self._download_base_session_artifacts(self.recce_cloud, org_id, project_id)

        # 3. Try to download existing recce_state, otherwise create new state
        try:
            logger.debug(f"Downloading recce_state for session {self.session_id}")
            state = self._download_session_recce_state(self.recce_cloud, org_id, project_id, self.session_id)
        except Exception as e:
            logger.debug(f"No existing recce_state found, creating new state: {e}")
            state = RecceState()

        if pr_url:
            pr_id = pr_url.rstrip("/").split("/")[-1]
            pull_request = PullRequestInfo(id=pr_id, url=pr_url)
            self.pr_info = pull_request
            if state.pull_request is None:
                state.pull_request = pull_request

        # Set artifacts regardless of whether we loaded existing state
        state.artifacts.base = base_artifacts
        state.artifacts.current = current_artifacts

        return state

    def _download_session_artifacts(self, recce_cloud, org_id: str, project_id: str, session_id: str) -> dict:
        """Download manifest and catalog for a session, return JSON data directly."""
        import requests

        # Get download URLs
        presigned_urls = recce_cloud.get_download_urls_by_session_id(org_id, project_id, session_id)

        artifacts = {}

        # Download manifest
        response = requests.get(presigned_urls["manifest_url"])
        if response.status_code == 200:
            artifacts["manifest"] = response.json()
        else:
            raise RecceException(f"Failed to download manifest for session {session_id}")

        # Download catalog
        response = requests.get(presigned_urls["catalog_url"])
        if response.status_code == 200:
            artifacts["catalog"] = response.json()
        else:
            raise RecceException(f"Failed to download catalog for session {session_id}")

        return artifacts

    def _download_session_recce_state(self, recce_cloud, org_id: str, project_id: str, session_id: str) -> RecceState:
        """Download recce_state for a session."""
        # Get download URLs (now includes recce_state_url)
        presigned_urls = recce_cloud.get_download_urls_by_session_id(org_id, project_id, session_id)
        recce_state_url = presigned_urls.get("recce_state_url")

        if not recce_state_url:
            raise RecceException(f"No recce_state_url found for session {session_id}")

        # Reuse the existing download method
        state = self._download_state_from_url(recce_state_url, SupportedFileTypes.FILE)

        if state is None:
            raise RecceException(f"Failed to download recce_state for session {session_id}")

        return state

    def _download_base_session_artifacts(self, recce_cloud, org_id: str, project_id: str) -> dict:
        """Download manifest and catalog for the base session, return JSON data directly."""
        import requests

        # Get download URLs for base session
        presigned_urls = recce_cloud.get_base_session_download_urls(org_id, project_id)

        artifacts = {}

        # Download manifest
        response = requests.get(presigned_urls["manifest_url"])
        if response.status_code == 200:
            artifacts["manifest"] = response.json()
        else:
            raise RecceException(f"Failed to download base session manifest for project {project_id}")

        # Download catalog
        response = requests.get(presigned_urls["catalog_url"])
        if response.status_code == 200:
            artifacts["catalog"] = response.json()
        else:
            raise RecceException(f"Failed to download base session catalog for project {project_id}")

        return artifacts

    def _export_state(self) -> Tuple[Union[str, None], str]:
        """
        Export state to Recce Cloud based on catalog type.

        Returns:
            str: A message indicating the result of the export operation.
            str: The etag of the exported state file (only used for GitHub).
        """
        logger.info("Store recce state to Recce Cloud")

        if self.catalog == "github":
            return self._export_state_to_github()
        elif self.catalog == "preview":
            return self._export_state_to_preview()
        elif self.catalog == "session":
            return self._export_state_to_session()
        else:
            raise RecceException(f"Unsupported catalog type: {self.catalog}")

    def _export_state_to_github(self) -> Tuple[Union[str, None], str]:
        """Export state to GitHub PR with metadata and etag."""
        if (self.pr_info is None) or (self.pr_info.id is None) or (self.pr_info.repository is None):
            raise RecceException("Cannot get the pull request information from GitHub.")

        # Generate metadata for GitHub only
        check_status = CheckDAO().status()
        metadata = {
            "total_checks": check_status.get("total", 0),
            "approved_checks": check_status.get("approved", 0),
        }

        # Upload to Cloud
        presigned_url = self.recce_cloud.get_presigned_url_by_github_repo(
            method=PresignedUrlMethod.UPLOAD,
            repository=self.pr_info.repository,
            artifact_name=RECCE_STATE_COMPRESSED_FILE,
            pr_id=self.pr_info.id,
            metadata=metadata,
        )
        message = self._upload_state_to_url(
            presigned_url=presigned_url,
            file_type=SupportedFileTypes.GZIP,
            password=self.cloud_options.get("password"),
            metadata=metadata,
        )

        # Get updated etag after upload
        metadata_response = self._get_metadata_from_recce_cloud()
        state_etag = metadata_response.get("etag") if metadata_response else None

        if message:
            logger.warning(message)
        return message, state_etag

    def _export_state_to_preview(self) -> Tuple[Union[str, None], None]:
        """Export state to preview share (no metadata or etag needed)."""
        share_id = self.cloud_options.get("share_id")
        presigned_url = self.recce_cloud.get_presigned_url_by_share_id(
            method=PresignedUrlMethod.UPLOAD,
            share_id=share_id,
            metadata=None,
        )
        message = self._upload_state_to_url(
            presigned_url=presigned_url, file_type=SupportedFileTypes.FILE, password=None, metadata=None
        )

        if message:
            logger.warning(message)
        return message, None

    def _export_state_to_session(self) -> Tuple[Union[str, None], None]:
        """Export state to session (upload recce_state with empty artifacts)."""
        if self.session_id is None:
            raise RecceException("Cannot export state to session. No session ID is provided.")

        # Get session information
        session = self.recce_cloud.get_session(self.session_id)
        org_id = session.get("org_id")
        project_id = session.get("project_id")

        if not org_id or not project_id:
            raise RecceException(f"Session {self.session_id} does not belong to a valid organization or project.")

        # Get upload URLs (now includes recce_state_url)
        presigned_urls = self.recce_cloud.get_upload_urls_by_session_id(org_id, project_id, self.session_id)
        recce_state_url = presigned_urls.get("recce_state_url")

        if not recce_state_url:
            raise RecceException(f"No recce_state_url found for session {self.session_id}")

        # Create a copy of the state with empty artifacts for upload
        upload_state = RecceState()
        upload_state.runs = self.state.runs.copy() if self.state.runs else []
        upload_state.checks = self.state.checks.copy() if self.state.checks else []
        # Keep artifacts empty (don't copy self.state.artifacts)

        # Upload the state with empty artifacts
        message = self._upload_state_to_url(
            presigned_url=recce_state_url,
            file_type=SupportedFileTypes.FILE,
            password=None,
            metadata=None,
            state=upload_state,
        )

        if message:
            logger.warning(message)
        else:
            # Notify Recce Cloud that the state has been uploaded if upload is successful
            self.recce_cloud.post_recce_state_uploaded_by_session_id(org_id, project_id, self.session_id)
        return message, None

    def _upload_state_to_url(
        self,
        presigned_url: str,
        file_type: SupportedFileTypes,
        password: str = None,
        metadata: dict = None,
        state: RecceState = None,
    ) -> Union[str, None]:
        """Upload state file to presigned URL."""
        import tempfile

        import requests

        # Use provided state or default to self.state
        upload_state = state or self.state

        # Prepare headers
        headers = {}
        if password:
            headers.update(s3_sse_c_headers(password))
        if metadata:
            headers["x-amz-tagging"] = urlencode(metadata)

        with tempfile.NamedTemporaryFile() as tmp:
            # Use the specified state to export to file
            json_data = upload_state.to_json()
            io = file_io_factory(file_type)
            io.write(tmp.name, json_data)

            with open(tmp.name, "rb") as fd:
                response = requests.put(presigned_url, data=fd.read(), headers=headers)

            if response.status_code not in [200, 204]:
                self.error_message = response.text
                return "Failed to upload the state file to Recce Cloud. Reason: " + response.text

        return None

    def check_conflict(self) -> bool:
        if self.catalog != "github":
            return False

        metadata = self._get_metadata_from_recce_cloud()
        if not metadata:
            return False

        state_etag = metadata.get("etag")
        return state_etag != self.state_etag


class RecceCloudStateManager:
    error_message: str
    hint_message: str

    # It is a class to upload, download and purge the state file on Recce Cloud.

    def __init__(self, cloud_options: Optional[Dict[str, str]] = None):
        self.cloud_options = cloud_options or {}
        self.pr_info = None
        self.error_message = None
        self.hint_message = None
        self.github_token = self.cloud_options.get("github_token")

        if not self.github_token:
            raise RecceException(RECCE_CLOUD_TOKEN_MISSING.error_message)
        self.pr_info = fetch_pr_metadata(cloud=True, github_token=self.github_token)
        if self.pr_info.id is None:
            raise RecceException("Cannot get the pull request information from GitHub.")

    def verify(self) -> bool:
        if self.github_token is None:
            self.error_message = RECCE_CLOUD_TOKEN_MISSING.error_message
            self.hint_message = RECCE_CLOUD_TOKEN_MISSING.hint_message
            return False
        if self.cloud_options.get("password") is None:
            self.error_message = RECCE_CLOUD_PASSWORD_MISSING.error_message
            self.hint_message = RECCE_CLOUD_PASSWORD_MISSING.hint_message
            return False
        return True

    @property
    def error_and_hint(self) -> (Union[str, None], Union[str, None]):
        return self.error_message, self.hint_message

    def _check_state_in_recce_cloud(self) -> bool:
        return RecceCloud(token=self.github_token).check_artifacts_exists(self.pr_info)

    def check_cloud_state_exists(self) -> bool:
        return self._check_state_in_recce_cloud()

    def _upload_state_to_recce_cloud(self, state: RecceState, metadata: dict = None) -> Union[str, None]:
        import tempfile

        import requests

        presigned_url = RecceCloud(token=self.github_token).get_presigned_url_by_github_repo(
            method=PresignedUrlMethod.UPLOAD,
            repository=self.pr_info.repository,
            artifact_name=RECCE_STATE_COMPRESSED_FILE,
            pr_id=self.pr_info.id,
            metadata=metadata,
        )

        compress_passwd = self.cloud_options.get("password")
        headers = s3_sse_c_headers(compress_passwd)
        with tempfile.NamedTemporaryFile() as tmp:
            state.to_file(tmp.name, file_type=SupportedFileTypes.GZIP)
            response = requests.put(presigned_url, data=open(tmp.name, "rb").read(), headers=headers)
            if response.status_code != 200:
                return f"Failed to upload the state file to Recce Cloud. Reason: {response.text}"
        return "The state file is uploaded to Recce Cloud."

    def upload_state_to_cloud(self, state: RecceState) -> Union[str, None]:
        if (self.pr_info is None) or (self.pr_info.id is None) or (self.pr_info.repository is None):
            raise RecceException("Cannot get the pull request information from GitHub.")

        checks = state.checks

        metadata = {
            "total_checks": len(checks),
            "approved_checks": len([c for c in checks if c.is_checked]),
        }

        return self._upload_state_to_recce_cloud(state, metadata)

    def _download_state_from_recce_cloud(self, filepath):
        import io

        import requests

        presigned_url = RecceCloud(token=self.github_token).get_presigned_url_by_github_repo(
            method=PresignedUrlMethod.DOWNLOAD,
            repository=self.pr_info.repository,
            artifact_name=RECCE_STATE_COMPRESSED_FILE,
            pr_id=self.pr_info.id,
        )

        password = self.cloud_options.get("password")
        if password is None:
            raise RecceException(RECCE_CLOUD_PASSWORD_MISSING.error_message)

        headers = s3_sse_c_headers(password)
        response = requests.get(presigned_url, headers=headers)

        if response.status_code != 200:
            raise RecceException(
                f"{response.status_code} Failed to download the state file from Recce Cloud. The password could be wrong."
            )

        byte_stream = io.BytesIO(response.content)
        gzip_io = file_io_factory(SupportedFileTypes.GZIP)
        decompressed_content = gzip_io.read_fileobj(byte_stream)

        dirs = os.path.dirname(filepath)
        if dirs:
            os.makedirs(dirs, exist_ok=True)
        with open(filepath, "wb") as f:
            f.write(decompressed_content)

    def download_state_from_cloud(self, filepath: str) -> Union[str, None]:
        if (self.pr_info is None) or (self.pr_info.id is None) or (self.pr_info.repository is None):
            raise RecceException("Cannot get the pull request information from GitHub.")

        logger.debug("Download state file from Recce Cloud...")
        return self._download_state_from_recce_cloud(filepath)

    def _purge_state_from_recce_cloud(self) -> (bool, str):
        try:
            RecceCloud(token=self.github_token).purge_artifacts(self.pr_info.repository, pr_id=self.pr_info.id)
        except RecceCloudException as e:
            return False, e.reason
        return True, None

    def purge_cloud_state(self) -> (bool, str):
        return self._purge_state_from_recce_cloud()


class RecceShareStateManager:
    error_message: str
    hint_message: str

    # It is a class to share state file on Recce Cloud.

    def __init__(self, auth_options: Optional[Dict[str, str]] = None):
        self.auth_options = auth_options or {}
        self.error_message = None
        self.hint_message = None

    def verify(self) -> bool:
        if get_recce_api_token() is None:
            self.error_message = RECCE_API_TOKEN_MISSING.error_message
            self.hint_message = RECCE_API_TOKEN_MISSING.hint_message
            return False
        return True

    @property
    def error_and_hint(self) -> (Union[str, None], Union[str, None]):
        return self.error_message, self.hint_message

    def share_state(self, file_name: str, state: RecceState) -> Dict:
        import tempfile

        with tempfile.NamedTemporaryFile() as tmp:
            state.to_file(tmp.name, file_type=SupportedFileTypes.FILE)
            response = RecceCloud(token=get_recce_api_token()).share_state(file_name, open(tmp.name, "rb"))
            return response



import pytest

from recce.tasks import ValueDiffDetailTask, ValueDiffTask


def test_value_diff(dbt_test_helper):
    csv_data_curr = """
        customer_id,name,age
        1,Alice,30
        2,Bob,25
        3,Charlie,35
        """

    csv_data_base = """
        customer_id,name,age
        1,Alice,35
        2,Bob,25
        3,Charlie,35
        """

    dbt_test_helper.create_model("customers", csv_data_base, csv_data_curr)
    params = dict(model="customers", primary_key=["customer_id"])
    task = ValueDiffTask(params)
    run_result = task.execute()
    assert len(run_result.data.columns) == 3
    assert len(run_result.data.data) == 3

    params = dict(model="customers", primary_key=["customer_id"])
    task = ValueDiffDetailTask(params)
    run_result = task.execute()
    assert len(run_result.columns) == 5
    assert len(run_result.data) == 2


def test_validator():
    from recce.tasks.valuediff import ValueDiffCheckValidator

    def validate(params: dict = {}, view_options: dict = {}):
        ValueDiffCheckValidator().validate(
            {
                "name": "test",
                "type": "value_diff",
                "params": params,
                "view_options": view_options,
            }
        )

    validate(
        {
            "model": "customers",
            "primary_key": "customer_id",
        }
    )
    validate(
        {
            "model": "customers",
            "primary_key": ["customer_id"],
        }
    )
    validate(
        {
            "model": "customers",
            "primary_key": ["customer_id"],
            "columns": ["name", "age"],
        }
    )

    with pytest.raises(ValueError):
        validate({})

    with pytest.raises(ValueError):
        validate(
            {
                "model": "customers",
            }
        )

    with pytest.raises(ValueError):
        validate(
            {
                "model": "customers",
                "primary_key": ["customer_id"],
                "columns": "name",
            }
        )
